"use strict";

var _get = function get(object, property, receiver) { if (object === null) object = Function.prototype; var desc = Object.getOwnPropertyDescriptor(object, property); if (desc === undefined) { var parent = Object.getPrototypeOf(object); if (parent === null) { return undefined; } else { return get(parent, property, receiver); } } else if ("value" in desc) { return desc.value; } else { var getter = desc.get; if (getter === undefined) { return undefined; } return getter.call(receiver); } };

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

function _asyncToGenerator(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

function _possibleConstructorReturn(self, call) { if (!self) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return call && (typeof call === "object" || typeof call === "function") ? call : self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function, not " + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

/**
 * @typedef {Object} ElementAttributes
 * @description attributes that can be set in a element using `addAttr` helper for Handlebar files.
 * */

var TemplateEngineHelpers = function () {
    function TemplateEngineHelpers() {
        _classCallCheck(this, TemplateEngineHelpers);
    }

    _createClass(TemplateEngineHelpers, null, [{
        key: "getHelpers",


        /**
         * @namespace
         * @description
         * Helpers written for handlebars. Can also be used by static references.
         * */
        value: function getHelpers(currentHelpers) {
            var if_helper = currentHelpers.if;
            return {
                /**
                 * @function
                 * @static
                 * @param {ElementAttributes} attributes - html element attributes.
                 * @param {String|Object} default_attributes - html element default attributes.
                 * @description
                 * attributes override default_attributes
                 * create attribute string and return which can be used to append
                 * @example
                 * let options = {
                * 	attributes : { id : 'element-id', className : 'css-class' }
                * }
                 * HandlebarUtils.execute(
                 * 	'<a {{addAttr attributes "{"href":"javascript:void(0)", "route":"true"}"}}></a>',
                 * 	options
                 * )
                 * //returns <a href="javascript:void(0)" route:"true" id="element-id" class:"css-class></a>
                 * */
                'addAttr': function addAttr(attributes, default_attributes) {
                    if (attributes || default_attributes) {
                        var attrData = [];
                        default_attributes = default_attributes && typeof default_attributes === "string" ? JSON.parse(default_attributes) : {};
                        var class_names = attributes ? attributes.className || attributes["class"] || "" : "";
                        class_names += default_attributes.hasOwnProperty("class") ? " " + default_attributes['class'] : "";
                        attributes = jQuery.extend({}, default_attributes, attributes);
                        attributes['class'] = class_names;
                        loop(attributes, function (key, value) {
                            if (key && value !== undefined) {
                                if (key === "style") {
                                    var css_values = [];
                                    loop(value, function (css_prop, css_value) {
                                        css_values.push(css_prop + ":" + css_value + ";");
                                    });
                                    value = css_values.join('');
                                }
                                if (typeof value === "string") {
                                    value = '"' + value.replace(/"/g, '\\"') + '"';
                                }
                                attrData.push(key + '=' + value);
                            }
                        });
                        return new window.Handlebars.SafeString(attrData.join(' '));
                    }
                },
                'selectedClass': function selectedClass(selected, selectedClassName) {
                    if (selected === true) {
                        return selectedClassName;
                    } else {
                        return '';
                    }
                },
                /**
                 * @function
                 * @static
                 * @param {String} template_name - precompiled handlebar template name.
                 * @param {Object} context - context with which the `template_name` hbs template to be called
                 * @return {String} `template_name` handlebar template contents after execution with provided context.
                 * @description
                 * includes one hbs template in another.
                 * @example
                 * //child HBS file - child-template.hbs
                 * `<div id="child">
                 *     {{name}}
                 * </div>`
                 *
                 * // main HBS file - main-template.hbs
                 * `<div id="main">
                 *     {{name}}
                 *     {{include 'child-template' child_context}}
                 * </div>`
                 *
                 * Handlebars.templates['main-template']({name:"MAIN", child_context:{name:"CHILD"}})
                 * // returns
                 * //<div id="main">
                 * //    MAIN
                 * //    <div id="child">
                 * //    	CHILD
                 * //	 </div>
                 * //</div>
                 * */
                'include': function include(template_name, context) {
                    return new window.Handlebars.SafeString(HandlebarUtils.execute(template_name, context));
                },
                /**
                 * @function
                 * @static
                 * @param {Number} param1 - left operand / lvalue
                 * @param {String} operation - can be one among "+", "-", "*", "/" and "%"
                 * @param {Number} param2 - right operand / rvalue
                 * @returns {Number} - result after operation between param1 and param2 with operation
                 * */
                'math': function math(param1, operation, param2) {
                    var lvalue = parseFloat(String(param1));
                    var rvalue = parseFloat(String(param2));
                    return {
                        "+": lvalue + rvalue,
                        "-": lvalue - rvalue,
                        "*": lvalue * rvalue,
                        "/": lvalue / rvalue,
                        "%": lvalue % rvalue
                    }[operation];
                },
                /**
                 * @function
                 * @static
                 * @param {String} i18nKey - application key that needs to be internationalized.
                 * @returns {String} internationalized application key's value
                 * @example
                 * // app.key.name = "APP Key"
                 * {{i18N 'app.key.name'}}	//returns "APP Key"
                 *
                 * // app.key.operation = "App Operation {0} by {1}"
                 * {{i18N 'app.key.operation' 'operationName' 'operatedBy'}}	//returns "App Operation operationName by operatedBy"
                 * */
                'i18n': function i18n(i18nKey) {
                    //~ (i18nKey, ...subs)
                    var subs = Array.prototype.slice.call(arguments);
                    subs = subs.splice(1, subs.length - 2);
                    return Adhara.i18n.get(i18nKey, subs);
                },
                /**
                 * @function
                 * @static
                 * @param {Object} object - object in which to lookup.
                 * @param {String} path - dot separated keys to be looked up for in depth.
                 * @description
                 * Wrapper for {@link getValueFromJSON}
                 * @returns {String|Number|Boolean|Object|Array} - Value for the key.
                 * */
                'get': function get(object, path) {
                    return getValueFromJSON(object, path);
                },
                /**
                 * @function
                 * @static
                 * @description
                 * Wrapper for {@link evaluateLogic}
                 * @returns
                 * handlebar block based on logic.
                 * */
                'if': function _if(param1, operator, param2, options) {
                    if ((typeof operator === "undefined" ? "undefined" : _typeof(operator)) === "object") {
                        return if_helper.call(this, param1, operator);
                    }
                    if (evaluateLogic(param1, operator, param2)) {
                        return options.fn(this);
                    } else {
                        return options.inverse(this);
                    }
                },
                /**
                 * @function
                 * @static
                 * @description
                 * If condition with multiple equations to evaluate.
                 * Takes numerous arguments which will be executed in pairs.
                 * @returns
                 * handlebar block based on logic.
                 * */
                'mIf': function mIf() {
                    var arr = [true, 'and', true];
                    var args_len = arguments.length - 1;
                    var options = arguments[args_len];
                    for (var i = 0; i < args_len; i++) {
                        if (i === 0) {
                            arr[0] = arguments[i];
                        } else if (i % 2 === 1) {
                            arr[1] = arguments[i];
                        } else {
                            arr[2] = arguments[i];
                            arr[0] = evaluateLogic(arr[0], arr[1], arr[2]);
                        }
                    }
                    if (arr[0]) {
                        return options.fn(this);
                    } else {
                        return options.inverse(this);
                    }
                },
                /**
                 * @function
                 * @static
                 * @description
                 * If condition with multiple equations to evaluate.
                 * Takes numerous arguments which will be executed in pairs.
                 * @returns {Boolean} result of equation created by provided params.
                 * @see evaluateLogic
                 * */
                'eIf': function eIf(param1, operator, param2) {
                    return evaluateLogic(param1, operator, param2);
                },
                /**
                 * @function
                 * @static
                 * @returns {String|Number|Boolean|Object|Array} - Value of the global letiable.
                 * */
                'global': function global(global_letiable) {
                    if (global_letiable.indexOf("Adhara.") === 0) {
                        return getValueFromJSON(Adhara, global_letiable.substring("Adhara.".length));
                    }
                    return getValueFromJSON(window, global_letiable);
                },
                'loop': function loop(looper, options) {
                    var structure = '';
                    for (var i = 0; i < looper.length; i++) {
                        structure += options.fn(looper[i]);
                    }
                    return structure;
                },
                /**
                 * @function
                 * @static
                 * @description Takes numerous {String} arguments and appends all those strings.
                 * @returns {String} arguments joined by ''.
                 * */
                'append': function append() {
                    var args = Array.prototype.slice.call(arguments);
                    args.pop();
                    return args.map(function (arg) {
                        if ((typeof arg === "undefined" ? "undefined" : _typeof(arg)) === "object") {
                            return JSON.stringify(arg);
                        }
                        return arg;
                    }).join('');
                },
                /**
                 * @function
                 * @static
                 * @description Takes a {String} and converts it to JSON object.
                 * @returns {Object}
                 * */
                'make_json': function make_json(str) {
                    return JSON.parse(str);
                },
                /**
                 * @function
                 * @static
                 * @description Takes a {String} and converts it to JSON object.
                 * @returns {Object}
                 * */
                'make_context': function make_context() {
                    var context = {};
                    for (var i = 0; i < arguments.length - 1; i += 2) {
                        context[arguments[i]] = arguments[i + 1];
                    }
                    return context;
                },
                /**
                 * @function
                 * @static
                 * @param {Function} fn
                 * @description Takes numerous and calls the fn with arguments from 2nd position since 1st argument is fn itself.
                 * @returns {String|Number|Boolean|Object|Array} content returned by fn
                 * */
                'fn_call': function fn_call(fn) {
                    return call_fn.apply(this, Array.prototype.slice.call(arguments).slice(0, -1));
                },
                /**
                 * @function
                 * @static
                 * @param {Object} looper
                 * @param {Object} options - Handlebars options that has access to blocks.
                 * @returns {String} Block string.
                 * @description blocks will be called with a context ~ {"key":k, "value":v}
                 * where `k` is key in the looper object and `v` is value corresponding to `k`.
                 * */
                'loopObject': function loopObject(looper, options) {
                    var structure = '';
                    loop(looper, function (key, value) {
                        structure += options.fn({ key: key, value: value });
                    });
                    return structure;
                },
                /**
                 * @function
                 * @static
                 * @param {Number} start
                 * @param {Number} step
                 * @param {Number} end
                 * @param {Object} options - Handlebars options that has access to blocks.
                 * @returns {String} Block string appended n times where n is the number of times step-looped.
                 * @description Block will be called with just a number {Number} as context which keeps increasing in steps, where the step height is equal to step param.
                 * */
                'iterate_range': function iterate_range(start, step, end, options) {
                    var str_buf = '';
                    for (var i = start; i <= end; i += step) {
                        str_buf += options.fn(i);
                    }
                    return str_buf;
                }
            };
        }
    }, {
        key: "registerHelpers",
        value: function registerHelpers() {
            window.Handlebars.registerHelper(TemplateEngineHelpers.getHelpers(window.Handlebars.helpers));
        }
    }]);

    return TemplateEngineHelpers;
}();

/**------------------------------------------------------------------------------------------------------------------**/


function handleForm(form) {
    if (form.submitting) {
        return;
    } else {
        form.submitting = true;
    }
    var hasFiles = !!form.querySelector('input[type="file"]');
    var apiData = void 0;
    if (hasFiles) {
        apiData = new FormData(form);
    } else {
        var formData = jQuery(form).serializeArray();
        apiData = {};
        jQuery.each(formData, function (i, fieldData) {
            apiData[fieldData.name] = fieldData.value;
        });
    }
    var format_fn = form.getAttribute('format-data');
    if (format_fn) {
        apiData = call_fn(format_fn, apiData);
    }
    if (apiData === false) {
        return;
    }
    RestAPI[form.getAttribute('api-method')]({
        url: form.action.split(window.location.host)[1],
        data: apiData,
        successMessage: form.getAttribute('success-message'),
        handleError: form.getAttribute('handle-error') !== "false",
        success: function success(data) {
            if (form.getAttribute('form-clear') === "true") {
                form.reset();
            }
            jQuery(form).trigger('success', data);
        },
        failure: function failure(message) {
            jQuery(form).trigger('failure', message);
        }
    });
}

function registerConfigUtils() {
    Adhara.configUtils = {
        getViewInstance: function getViewInstance(entity_config) {
            try {
                return Adhara.getView(entity_config.view);
            } catch (e) {
                return false;
            }
        },
        getProcessor: function getProcessor(entity_config) {
            return entity_config.processor || Processor.fallback;
        },
        getController: function getController(entity_config) {
            return entity_config.controller || Adhara.restAPI;
        },
        getDataConfig: function getDataConfig(entity_config) {
            return entity_config.data_config;
        },
        getBlobClass: function getBlobClass(entity_config) {
            return entity_config.data_config.blob || entity_config.blob || DataBlob;
        },
        getByDataConfig: function getByDataConfig(data_config) {
            var res = [];
            var app_config = Adhara.app.config;
            for (var entity_name in app_config) {
                if (app_config.hasOwnProperty(entity_name)) {
                    var entity_config = Adhara.app.getEntityConfig(entity_name);
                    if (data_config._url === Adhara.configUtils.getDataConfig(entity_config)._url && (!data_config.blob || data_config.blob === Adhara.configUtils.getBlobClass(entity_config)) || JSON.stringify(data_config) === JSON.stringify(entity_config.data_config)) {
                        res.push(entity_config);
                    }
                }
            }
            return res;
        }
    };
}

function registerAdharaUtils() {
    //config utils
    registerConfigUtils();
    //Register templateEngine helpers
    Adhara.templateEngine.helpersHandler.registerHelpers();
    //Form listeners
    jQuery(document).on('submit', 'form.api-form', function (event) {
        event.preventDefault();
        handleForm(this);
    });
    //Form listeners
    jQuery(document).on('success', 'form.dialog-form', function () /*e, d*/{
        this.close.click();
    });
}

/**
 * @function
 * @global
 * @param {String|Number|Boolean} param1
 * @param {String} operator - Operator can be one among these.
 * "in", "not_in", "contains", "not_contains", "has", "||", "&&", "|", "&", "==", "===", "!=", "!==", ">", "<", ">=", "<=", "equals", "and", "or".
 * @param {String|Number|Boolean} param2
 * @Returns {Boolean}
 *
 * "in" - param2 must be an {Array}. Whether param1's existence in that array will be checked for
 *
 * "not_in" - param2 must be an {Array} and param1's non-existence in that array will be checked for
 *
 * "contains" - param1 must be an {Array} and param2's existence in that array will be checked for
 *
 * "not_contains" - param1 must be an {Array} and param2's non-existence in that array will be checked for
 *
 * "has" - param1 must be an {Object}. Whether param2 is a key of param1
 *
 * All other operations will be applied as javaScript evaluates them.
 * `"equals" is equivalent to "=="`,
 * `"and" is equivalent to "&&"`
 * and `"or" is equivalent to "||"`.
 *
 * */
function evaluateLogic(param1, operator, param2) {
    if (operator === "in") {
        return param2 && param2.indexOf(param1) !== -1;
    } else if (operator === "not_in") {
        return param2 && param2.indexOf(param1) === -1;
    } else if (operator === "contains") {
        return param1 && param1.indexOf(param2) !== -1;
    } else if (operator === "not_contains") {
        return param1 && param1.indexOf(param2) === -1;
    } else if (operator === "has") {
        return param1 && param1.hasOwnProperty(param2);
    } else {
        return {
            "||": param1 || param2,
            "&&": param1 && param2,
            "|": param1 | param2,
            "&": param1 & param2,
            "==": param1 == param2,
            "===": param1 === param2,
            "!=": param1 != param2,
            "!==": param1 !== param2,
            ">": param1 > param2,
            "<": param1 < param2,
            ">=": param1 >= param2,
            "<=": param1 <= param2,
            "equals": param1 == param2,
            "and": param1 && param2,
            "or": param1 || param2
        }[operator];
    }
}

/**
 * @function
 * @global
 * @param {Object} object - object in which to lookup.
 * @param {String} path - dot separated keys to be looked up for in depth.
 * @param {String} [identifier=null] - token that helps to lookup inside arrays.
 * @returns {String|Number|Boolean|Object|Array} - Value for the key.
 * @description
 * Looks up a JSON object in depth and returns the value for the key provided.
 * @example
 * let obj = {
 *	task: {
 *		id: "12341234",
 * 		status: {
 * 			color: "#fff"
 * 			}
 *		}
 *	};
 * getValueFromJSON(obj, "task.status.color");	//returns "#fff"
 *
 * let objX = {
 *	tasks: [
 *		{
 *			id: "12341234",
 * 			status: {
 * 				color: "#fff"
 * 				}
 *			},
 *		{
 *			id: "55424",
 * 			status: {
 * 				color: "#f5f5f5"
 * 				}
 *			},
 *		{
 *			id: "90898080",
 * 			status: {
 * 				color: "#eee"
 * 				}
 *			},
 *		]
 *	};
 * getValueFromJSON(objX, "task[1].status.color");	//returns "#f5f5f5"
 * getValueFromJSON(objX, "task[$, ].status.color", "$");	//returns "#fff, #f5f5f5, #eee"
 * // Elaborately,
 * // In `[$, ]`		--split into identifier and separator
 * // '$'=identifier(from params).
 * // Remaining part inside [] is the separator i.e., `', '` will be used to join the results from the array
 * */
function getValueFromJSON(object, path, identifier) {
    try {
        if (!path) {
            return object;
        }
        var keys = path.split('.');
        for (var i = 0; i < keys.length; i++) {
            var key_in = keys[i];
            if (key_in.indexOf('[') !== -1) {
                var arr = key_in.split('[');
                var arrName = arr[0];
                var index = arr[1].split(']')[0];
                if (isNaN(index)) {
                    var rite = object[arrName];
                    var separator = index.substring(identifier.length, index.length);
                    var formattedVal = '';
                    for (var j = 0; j < rite.length; j++) {
                        if (j > 0) {
                            formattedVal += separator;
                        }
                        formattedVal += getValueFromJSON(rite[j], path.split('.').splice(i + 1).join('.'), identifier);
                    }
                    object = formattedVal;
                } else {
                    object = object[arrName][index];
                }
                break;
            } else {
                object = object[key_in];
            }
        }
        return object;
    } catch (e) {
        return undefined;
    }
}

/**
 * @function
 * @global
 * @param {Object} object - object to set value to
 * @param {String} path - dot separated path.
 * @param {String|Number|Boolean|Object|Array} value - value to be set fot the given key
 * @description
 * Sets value to provided object at specified depth in the path be using dot separators.
 * Creates objects at depths if no object already exists at specified path.
 * @example
 * let kit = {'has_bat': true}
 * setValueToJson(kit, 'has_ball', true);	//let kit = {'has_bat': true, 'has_ball': true}
 * setValueToJson(kit, 'has_bat', false);	//let kit = {'has_bat': false, 'has_ball': true}
 *
 * let aeroplane = {'name': 'B747'}
 * setValueToJson(aeroplane, 'tank.capacity', '4000');	//aeroplane = {'name': 'B747', 'tank': { 'capacity': '4000' }}
 * setValueToJson(aeroplane, 'tank.shape', 'amoeba');	//aeroplane = {'name': 'B747', 'tank': { 'capacity': '4000', 'shape': 'amoeba' }}
 * */
function setValueToJson(object, path, value) {
    var keys = path.split('.');
    loop(keys, function (i, key) {
        if (i + 1 < keys.length) {
            if (!object.hasOwnProperty(key)) {
                object[key] = {};
            }
            object = object[key];
        } else {
            object[key] = value;
        }
    });
}

/**
 * @typedef {Object|Array|String|Boolean|Number|null|undefined} _MultiParams - numerous arguments of any kind
 * @description
 * A function if accepts multiple params, can use this as param type.
 * @example
 * //def
 * function mpFn({String} param1, {_MultiParams} m_params){}
 * //call
 * mpFn("Str_param1", "p1", "p2", 1, {'key':'val'}, ['m1', 'm2', ...], ...);
 * //def2
 * function mp2Fn({String} param1, {_MultiParams} m_params, {String} param2){}
 * //call2
 * mp2Fn("Str_param1", "p1", "p2", 1, {'key':'val'}, ['m1', 'm2', ...], ..., "StrX_param2");
 *
 * */
/**
 * @function
 * @global
 * @param {Function|String} fn - function to be called or global path to a function
 * @param {_MultiParams} params - any params that to be passed to function
 * @description
 * calls the function with params passed
 * @example
 * let gv = {}; //global_letiable
 * gv.sample_fn = function(param1, param2, param3){
 * 	//Do Something...
 * 	return "Hello from SampleFN"
 * };
 * call_fn('gv.sample_fn', param1, param2, param3);	//returns "Hello from SampleFN"
 * */
function call_fn(fn) {
    if (!fn) {
        return;
    }
    var args = Array.prototype.slice.call(arguments);
    args.splice(0, 1);
    if (typeof fn === "function") {
        return fn.apply(fn, args);
    } else {
        try {
            fn = getValueFromJSON(window, fn);
            if (fn && fn !== window) {
                return fn.apply(window[fn], args);
            }
        } catch (e) {
            if (e.name === "TypeError" && e.message.indexOf("is not a function") !== -1) {
                throw new Error("error executing function : " + fn);
            } else {
                throw e;
            }
        }
    }
}

/**
 * Can be used if required to call a function, and if it is unavailable, call a default fn
 * Usage : call_fn_or_def(fn, arg1, arg2, arg3, def_fn)
 * Result : fn(arg1, arg2, arg3) if fn is not undefined else def_fn(arg1, arg2, arg3)
 * */

/**
 * @function
 * @global
 * @param {Function|String} fn - function to be called or global path to a function
 * @param {_MultiParams} params - any params that to be passed to function
 * @param {Function|String} fn - function to be called or global path to a function
 * @description
 * calls the function with params passed
 * @example
 * let gv = {}; //global_letiable
 * gv.sample_fn = function(param1, param2, param3){
 * 	//Do Something...
 * 	return "Hello from SampleFN"
 * };
 * call_fn('gv.sample_fn', param1, param2, param3, defaultFunction);	//returns "Hello from SampleFN"
 * */
function call_fn_or_def(fn) {
    var args = Array.prototype.slice.call(arguments);
    var def_fn = args.pop();
    if (!(typeof def_fn === "function" || typeof window[def_fn] === "function")) {
        args.push(def_fn);
        def_fn = undefined;
    }
    if (fn) {
        return call_fn.apply(call_fn, args);
    } else if (def_fn) {
        args[0] = def_fn;
        return call_fn.apply(call_fn, args);
    }
}

/**
 * Loop - looper iterator function
 * @callback LoopIteratorCallback
 * @param {String|Number|RegExp} key - key will be string in case of an object and Number will be an Array while looping through an Array.
 * @param {*} value - Value of iterable for key.
 * @returns {Boolean}
 * return `false` to stop iteration.
 * */
/**
 * @function
 * @global
 * @param {Object|Array} object - Iterable.
 * @param {LoopIteratorCallback} cbk - callback for each value in iterable.
 * @param {Boolean} [reverseIterate=false] - iterate in reverse if it is an array.
 * */
function loop(object, cbk, reverseIterate) {
    var i = void 0,
        loop_size = void 0;
    if (object instanceof Array) {
        loop_size = object.length;
        if (reverseIterate) {
            for (i = loop_size - 1; i >= 0; i--) {
                if (cbk(i, object[i]) === false) {
                    break;
                }
            }
        } else {
            for (i = 0; i < loop_size; i++) {
                if (cbk(i, object[i]) === false) {
                    break;
                }
            }
        }
    } else {
        var keys = Object.keys(object);
        loop_size = keys.length;
        for (i = 0; i < loop_size; i++) {
            if (cbk(keys[i], object[keys[i]]) === false) {
                break;
            }
        }
    }
}

/**
 * @typedef {String} HandlebarTemplate
 * @description handlebar template can be an inline template or a pre compiled template that is created in a HBS file.
 * */

/**
 * @namespace
 * @description
 * HandlebarUtils is a set of utilities related to handlebars
 * */
var HandlebarUtils = {};
(function () {
    var preCompiledCache = {};

    /**
     * @function
     * @static
     * @param {HandlebarTemplate} template_or_template_string - Handlebar template name or template string.
     * @param {Object} context - content to be passed t provided handlebar template/string.
     * @param {Boolean} [cache=true] - whether to cache if the provided template is a Handlebar string template.
     * @returns {String} compiled adn executed handlebar template
     * @example
     * //using template
     * HandlebarUtils.execute('template-file-name', {context_key1: "value1", ...});
     *
     * //using template string
     * HandlebarUtils.execute('Hello {{name}}', {name: "Adhara"});	//returns "Hello Adhara"
     * */
    HandlebarUtils.execute = function (template_or_template_string, context, cache) {
        //Check if it is a pre compiled hbs template
        var template = window.Handlebars.templates[template_or_template_string];
        if (!template && window.Handlebars.hasOwnProperty("compile")) {
            //If template not found in precompiled hbs list => template is a handlebar string template
            // check if cacheable
            if (cache !== false) {
                //check if the template is already cached.
                template = preCompiledCache[template_or_template_string];
                if (!template) {
                    //else compile, store it in cache and proceed to execution
                    template = preCompiledCache[template_or_template_string] = window.Handlebars.compile(template_or_template_string);
                }
            } else {
                //else compile and proceed to execution
                template = window.Handlebars.compile(template_or_template_string);
            }
        }
        //execute with the provided context and return the output content...
        return template(context);
    };
})();

var Internationalize = function () {

    /**
     * @constructor
     * @param {Object<String, String>} key_map - i18n key map
     * */
    function Internationalize(key_map) {
        _classCallCheck(this, Internationalize);

        this.key_map = key_map;
    }

    /**
     * @instance
     * @function
     * @param {String} key - key
     * @param {Array<String>} subs - substitutes
     * @param {String} default_value - will be returned if key is not availalbe in the keymap
     * */


    _createClass(Internationalize, [{
        key: "getValue",
        value: function getValue(key, subs, default_value) {
            var value = this.key_map[key] || getValueFromJSON(this.key_map, key);
            if (!value) {
                return default_value;
            }
            subs = subs || [];
            var placeholders = value.match(/{[0-9]+}/g) || [];
            for (var i = 0; i < placeholders.length; i++) {
                var sub = subs[i] || "";
                try {
                    if (sub.indexOf('.') !== -1) {
                        sub = this.get(sub);
                    }
                } catch (e) {/*Do Nothing*/}
                value = value.replace(new RegExp("\\{" + i + "\\}", "g"), sub);
            }
            return value.trim();
        }

        /**
         * @instance
         * @function
         * @param {String} key - key
         * @param {Array<String>} [subs=[]] - substitutes
         * */

    }, {
        key: "get",
        value: function get(key, subs) {
            return this.getValue(key, subs, key);
        }
    }]);

    return Internationalize;
}();

// Enable multi extend functionality https://stackoverflow.com/a/45332959


function MutateViews(baseClass) {
    for (var _len = arguments.length, mixins = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
        mixins[_key - 1] = arguments[_key];
    }

    var base = function (_baseClass) {
        _inherits(base, _baseClass);

        function base() {
            var _ref;

            for (var _len2 = arguments.length, args = Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
                args[_key2] = arguments[_key2];
            }

            _classCallCheck(this, base);

            var _this = _possibleConstructorReturn(this, (_ref = base.__proto__ || Object.getPrototypeOf(base)).call.apply(_ref, [this].concat(args)));

            mixins.forEach(function (mixin) {
                copyProps(_this, new (Function.prototype.bind.apply(mixin, [null].concat(args)))());
            });
            return _this;
        }

        return base;
    }(baseClass);

    function copyProps(target, source) {
        // this function copies all properties and symbols, filtering out some special ones
        //TODO "Object.getOwnPropertyNames" and "Object.getOwnPropertySymbols" Doesn't work in IE even after transpilation to ES5...
        Object.getOwnPropertyNames(source).concat(Object.getOwnPropertySymbols(source)).forEach(function (prop) {
            if (!prop.match(/^(?:constructor|prototype|arguments|caller|name|bind|call|apply|toString|length)$/)) Object.defineProperty(target, prop, Object.getOwnPropertyDescriptor(source, prop));
        });
    }
    mixins.forEach(function (mixin) {
        // outside contructor() to allow aggregation(A,B,C).staticFunction() to be called etc.
        copyProps(base.prototype, mixin.prototype);
        copyProps(base, mixin);
    });
    return base;
}
/**------------------------------------------------------------------------------------------------------------------**/
/**
 * @namespace
 * @description
 * Client routing logic.
 * All the client URL pattern's should be registered with the router.
 *
 * On visiting a URL, if called {@link AdharaRouter.route}, it calls the registered view function,
 * which takes care of rendering the page.
 *
 * Any HTML element can be used as a router. For this purpose,
 *
 *
 * > Add this attribute `href="/url/to/be/navigated/to"` and this property `route` to any html element to make behave as a router.
 *
 * > By default routing happening this way will execute the view function even if the current URL is same as provided href.
 *
 * > In order to disable this behaviour, provide the attribute `data-force="false"`
 *
 * > In order to use {@link AdharaRouter.goBack} functionality provide `data-force="true"`. See example below.
 *
 * @example
 * //Registering
 * AdharaRouter.register("^/adhara/{{instance_id}}([0-9]+)/{{tab}}(details|tokens|history)$", function(instance_id, tab){
 *		console.log(instance_id, tab);
 *	    AdharaRouter.getPathParam("instance_id");  //Returns same as instance_id
 *	    AdharaRouter.getPathParam("tab");  //Returns same as tab
 * });
 *
 * //Navigating
 * AdharaRouter.navigateTo("/adhara/123412341234/details");
 *
 * //Routing - In case if URL is already set in address bar and not via Router, call this function to execute the registered view funciton.
 * this.route();
 *
 * //HTML Example
 *
 * <a href="/adhara/123412341234/tokens" />
 *
 * <button href="/adhara/123412341234/tokens" data-force="false">tokens</button>
 *
 * <div href="/adhara/123412341234/details" data-back="true"></div>
 *
 * */
var AdharaRouter = null;

(function () {

    "use strict";

    /**
     * @private
     * @member {Object<String, Object>}
     * @description
     * Stores all the registered URL's along with other view parameters.
     * */

    var registeredUrlPatterns = {};

    /**
     * @private
     * @member {Object<String, String>}
     * @description
     * Stores the current URL's search query parameters.
     * */
    var queryParams = {};

    /**
     * @private
     * @member {Object<String, String>}
     * @description
     * Stores the current URL's path variables.
     * */
    var pathParams = {};

    /**
     * @private
     * @member {Array<String>}
     * @description
     * Stores list of visited URL's
     * */
    var historyStack = [];

    /**
     * @private
     * @member {String|undefined}
     * @description
     * Stores current page URL.
     * */
    var currentUrl = undefined;

    /**
     * @private
     * @member {RouterURLConf}
     * @description
     * Stores route which matches with the current URL against registered URL patterns.
     * */
    var currentRoute = undefined;

    /**
     * @private
     * @member {Object<String, Function|undefined>}
     * @description
     * Stores listeners that will be called on routing.
     * */
    var listeners = {};

    /**
     * @typedef {Function} AdharaRouterMiddleware
     * @param {Object} params - url parameters
     * @param {String } params.view_name - name of the page that is being routed to
     * @param {String} params.path - path that is being routed to
     * @param {Object} params.query_params - url query parameters
     * @param {Object} params.path_params - url path parameters
     * @param {Function} route - Proceed with routing after making necessary checks in middleware
     * */

    /**
     * @private
     * @member {Array<AdharaRouterMiddleware>}
     * @description
     * Stores middlewares that will be called on routing.
     * */
    var middlewares = [];

    /**
     * @function
     * @private
     * @returns {String} Current window URL. IE compatibility provided by Hostory.js (protocol://domain/path?search_query).
     * */
    function getFullUrl() {
        return window.location.href;
        // return History.getState().cleanUrl;
    }

    /**
     * @function
     * @private
     * @returns {String} Full URL without search query (protocol://domain/path).
     * */
    function getBaseUrl() {
        return getFullUrl().split('?')[0];
    }

    /**
     * @function
     * @private
     * @returns {String} URL Path (path).
     * */
    function getPathName() {
        return AdharaRouter.transformURL(window.location.pathname);
    }

    /**
     * @function
     * @private
     * @returns {String} URL Path with search query (/path?search_query).
     * */
    function getFullPath() {
        return getPathName() + window.location.search + window.location.hash;
        // return "/"+getFullUrl().split('://')[1].substring(window.location.host.length+1);
    }

    /**
     * @function
     * @private
     * @returns {String} Search query (search_query).
     * */
    function getSearchString() {
        return window.location.search.substring(1);
    }

    /**
     * @function
     * @private
     * @returns {String} Hash (text after `#` from the url).
     * */
    function getHash() {
        return window.location.hash.substring(1);
    }

    /**
     * @function
     * @private
     * @param {String} new_path - URL path.
     * @returns {Boolean} Whether new_path matches with current URL path.
     * */
    function isCurrentPath(new_path) {
        function stripSlash(path) {
            if (path.indexOf('/') === 0) {
                return path.substring(1);
            }
            return path;
        }
        return stripSlash(getFullPath()) === stripSlash(new_path);
    }

    function callMiddlewares(params, proceed) {
        var i = 0;
        (function _proceed() {
            var middleware_fn = middlewares[i++];
            if (middleware_fn) {
                call_fn(middleware_fn, params, _proceed);
            } else {
                proceed();
            }
        })();
    }

    /**
     * @function
     * @private
     * @description matches the current URL and returns the configuration, path and params
     * @returns Object
     * */
    function matchUrl() {
        var path = getPathName();
        var matchFound = false;
        for (var regex in registeredUrlPatterns) {
            if (registeredUrlPatterns.hasOwnProperty(regex) && !matchFound) {
                var formed_regex = new RegExp(regex);
                if (formed_regex.test(path)) {
                    matchFound = true;
                    var opts = registeredUrlPatterns[regex];
                    var params = formed_regex.exec(path);
                    return { opts: opts, path: path, params: params };
                }
            }
        }
    }

    /**
     * @function
     * @private
     * @description
     * Routes to the current URL.
     * Looks up in the registered URL patterns for a match to current URL.
     * If match found, view function will be called with regex path matches in the matched order and query param's as the last argument.
     * Current view name will be set to the view name configured against view URL.
     * @returns {Boolean} Whether any view function is found or not.
     * */
    function matchAndCall() {
        var _this2 = this;

        var matchOptions = matchUrl();
        if (matchOptions) {
            var opts = matchOptions.opts,
                path = matchOptions.path,
                params = matchOptions.params;

            params.splice(0, 1);
            if (opts && opts.fn) {
                var _pathParams = {};
                var _iteratorNormalCompletion = true;
                var _didIteratorError = false;
                var _iteratorError = undefined;

                try {
                    for (var _iterator = params.entries()[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
                        var _ref2 = _step.value;

                        var _ref3 = _slicedToArray(_ref2, 2);

                        var index = _ref3[0];
                        var param = _ref3[1];

                        _pathParams[opts.path_param_keys[index]] = param;
                    }
                } catch (err) {
                    _didIteratorError = true;
                    _iteratorError = err;
                } finally {
                    try {
                        if (!_iteratorNormalCompletion && _iterator.return) {
                            _iterator.return();
                        }
                    } finally {
                        if (_didIteratorError) {
                            throw _iteratorError;
                        }
                    }
                }

                currentRoute = opts;
                //Setting current routing params...
                pathParams = _pathParams;
                currentUrl = getFullUrl();
                fetchQueryParams();

                callMiddlewares({
                    view_name: opts.view_name,
                    path: path,
                    query_params: getQueryParams(),
                    path_params: _pathParams
                }, function () {
                    params.push(queryParams);
                    if (opts.fn.constructor instanceof AdharaView.constructor) {
                        Adhara.onRoute(opts.fn, params);
                    } else {
                        opts.fn.apply(_this2, params);
                    }
                });
            }
        }
        return !!matchOptions;
    }

    var curr_path = void 0;
    function updateHistoryStack() {
        if (curr_path) {
            historyStack.push(curr_path);
        }
        curr_path = getFullPath();
    }

    /**
     * @private
     * @member {Boolean}
     * @description
     * This flag will be set to true when set to true, {@link Router.route} will not call the view function.
     * */
    var settingURL = false;

    /**
     * @function
     * @private
     * @description
     * Updates the {AdharaRouter~queryParams} with the current URL parameters
     * */
    function updateParams() {
        if (getFullUrl() !== currentUrl) {
            currentUrl = getFullUrl();
            fetchQueryParams();
        }
    }

    /**
     * @function
     * @private
     * @description
     * Fetches the search query from current URL and transforms it to query param's object.
     * @returns {Object<String, String>} The search query will be decoded and returned as an object.
     * */
    function getQueryParams() {
        var qp = {};
        if (getSearchString()) {
            loop(getSearchString().split('&'), function (i, paramPair) {
                paramPair = paramPair.split('=');
                qp[decodeURIComponent(paramPair[0])] = decodeURIComponent(paramPair[1]);
            });
        } else {
            qp = {};
        }
        return qp;
    }

    /**
     * @function
     * @private
     * @description
     * Fetches the search query from current URL and transforms it to query param's.
     * The search query will be decoded and stored.
     * */
    function fetchQueryParams() {
        queryParams = getQueryParams();
    }

    /**
     * @function
     * @private
     * @returns {String} URL constructed by current base URL and current Query Parameters.
     * */
    function generateCurrentUrl() {
        var url = getPathName();
        var params = [];
        loop(queryParams, function (key, value) {
            params.push(key + "=" + value);
        });
        if (params.length) {
            url += "?" + params.join("&");
        }
        return url;
    }

    /**
     * @function
     * @private
     * @param {RouteType} [route_type=this.routeTypes.NAVIGATE] - How to route.
     * @param {Object} route_options - Route options.
     * @param {boolean} [route_options.force=false] - Whether to force navigate URL or not. Will call view function even if current URL is same. Applies for {@link RouteType.NAVIGATE}.
     * @see {@link RouteType}
     * */
    function _route(route_type, route_options) {
        if (!route_type || route_type === this.RouteTypes.SET) {
            this.setURL(generateCurrentUrl());
        } else if (route_type === this.RouteTypes.NAVIGATE) {
            this.navigateTo(generateCurrentUrl(), route_options && route_options.force || false);
        } else if (route_type === this.RouteTypes.OVERRIDE) {
            this.overrideURL(generateCurrentUrl());
        } else if (route_type === this.RouteTypes.UPDATE) {
            this.updateURL(generateCurrentUrl());
        }
    }

    var STATE_KEY = "__adhara_router__";

    var Router = function () {
        function Router() {
            _classCallCheck(this, Router);
        }

        _createClass(Router, null, [{
            key: "transformURL",


            /**
             * @function
             * @private
             * @param {String} url - path
             * @description
             * modified the URL and returns the new value. Default transformer just returns the passed url/path.
             * */
            value: function transformURL(url) {
                return url;
            }

            /**
             * @callback ViewFunction
             * @param {String} regexMatchedParams - matched regex params.
             * @param {String} searchQuery - search query parameters converted to object form.
             * @description
             * View callback function.
             * Parameters passed will be the regex matches and last parameter will be searchQueryParameters.
             * In case of more than 1 path regex match, ViewFunction will be called back with (match1, match2, ..., searchQuery)
             * */

            /**
             * @function
             * @static
             * @param {String} pattern - URL Pattern that is to be registered.
             * @param {String} view_name - Name of the view that is mapped to this URL.
             * @param {ViewFunction|Adhara} fn - View function that will be called when the pattern matches window URL.
             * */

        }, {
            key: "register_one",
            value: function register_one(pattern, view_name, fn) {
                var path_param_keys = [];
                var regex = /{{([a-zA-Z$_][a-zA-Z0-9$_]*)}}/g;
                var match = regex.exec(pattern);
                while (match != null) {
                    path_param_keys.push(match[1]);
                    match = regex.exec(pattern);
                }
                pattern = pattern.replace(new RegExp("\\{\\{[a-zA-Z$_][a-zA-Z0-9$_]*\\}\\}", "g"), '');
                //Removing all {{XYZ}} iff, XYZ matches first character as an alphabet ot $ or _

                pattern = "^" + this.transformURL(pattern.substring(1));
                pattern = pattern.replace(/[?]/g, '\\?'); //Converting ? to \? as RegExp(pattern) dosen't handle that
                registeredUrlPatterns[pattern] = { view_name: view_name, fn: fn, path_param_keys: path_param_keys };
            }

            /**
             * @function
             * @static
             * @param {RegExp} regex - URL Pattern that is to be unregistered.
             * @description
             * Once unregistered, this URL regex will be removed from registered URL patterns and on hitting that URL,
             * will be routed to a 404.
             * */

        }, {
            key: "deRegister",
            value: function deRegister(regex) {
                registeredUrlPatterns[regex] = undefined;
            }

            /**
             * Bulk url conf
             * @typedef {Object} RouterURLConf
             * @property {string} url - url regex
             * @property {String} view_name - view name
             * @property {Object} path_params - Path params
             * @property {Function|class} view - view
             * */

            /**
             * @function
             * @static
             * @param {Array<RouterURLConf>} list - list of url configurations
             * @description
             * Register's urls in the list iteratively...
             * */

        }, {
            key: "register",
            value: function register(list) {
                var _iteratorNormalCompletion2 = true;
                var _didIteratorError2 = false;
                var _iteratorError2 = undefined;

                try {
                    for (var _iterator2 = list[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
                        var conf = _step2.value;

                        this.register_one(conf.url, conf.view_name, conf.view);
                    }
                } catch (err) {
                    _didIteratorError2 = true;
                    _iteratorError2 = err;
                } finally {
                    try {
                        if (!_iteratorNormalCompletion2 && _iterator2.return) {
                            _iterator2.return();
                        }
                    } finally {
                        if (_didIteratorError2) {
                            throw _iteratorError2;
                        }
                    }
                }
            }

            /**
             * @typedef {Object} AdharaRouterConfiguration - Adhara router configuration
             * @property {Array<RouterURLConf>} routes - Route configurations
             * @property {Object<String, Function>} on_route_listeners - On Route listeners
             * @property {Array<AdharaRouterMiddleware>} middlewares - Middleware functions
             * */

            /**
             * @function
             * @static
             * @param {AdharaRouterConfiguration} router_configuration
             * */

        }, {
            key: "configure",
            value: function configure(router_configuration) {
                if (router_configuration.routes) {
                    Router.register(router_configuration.routes);
                }
                if (router_configuration.on_route_listeners) {
                    for (var listener_name in router_configuration.on_route_listeners) {
                        if (router_configuration.on_route_listeners.hasOwnProperty(listener_name)) {
                            Router.onRoute(listener_name, router_configuration.on_route_listeners[listener_name]);
                        }
                    }
                }
                if (router_configuration.middlewares) {
                    var _iteratorNormalCompletion3 = true;
                    var _didIteratorError3 = false;
                    var _iteratorError3 = undefined;

                    try {
                        for (var _iterator3 = router_configuration.middlewares[Symbol.iterator](), _step3; !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
                            var middleware_fn = _step3.value;

                            Router.registerMiddleware(middleware_fn);
                        }
                    } catch (err) {
                        _didIteratorError3 = true;
                        _iteratorError3 = err;
                    } finally {
                        try {
                            if (!_iteratorNormalCompletion3 && _iterator3.return) {
                                _iterator3.return();
                            }
                        } finally {
                            if (_didIteratorError3) {
                                throw _iteratorError3;
                            }
                        }
                    }
                }
            }

            /**
             * @function
             * @static
             * @param {String} listener_name - A unique name for the listener that will be useful for unregistering.
             * @param {Function} listener_fn - Listener function that to be called on routing. No arguments passed.
             * @description
             * All registered listeners will be called whenever routing happens via AdharaRouter
             * */

        }, {
            key: "onRoute",
            value: function onRoute(listener_name, listener_fn) {
                listeners[listener_name] = listener_fn;
            }

            /**
             * @function
             * @static
             * @param {String} listener_name - A unique name for the listener that will be useful for unregistering.
             * @description
             * De-reigsters the listener registered with this name.
             * */

        }, {
            key: "offRouteListener",
            value: function offRouteListener(listener_name) {
                listeners[listener_name] = undefined;
            }

            /**
             * @function
             * @static
             * @param {Function} middleware_fn - middleware function to be used
             * @description Stores all middleware's and calls then in the order of registration
             * */

        }, {
            key: "registerMiddleware",
            value: function registerMiddleware(middleware_fn) {
                if (typeof middleware_fn === "function") {
                    middlewares.push(middleware_fn);
                }
            }

            /**
             * @function
             * @static
             * @description
             * Routes to the current URL.
             * Looks up in the registered URL patterns for a match to current URL.
             * If match found, view function will be called with regex path matches in the matched order and query param's as the last argument.
             * Current view name will be set to the view name configured against view URL.
             * */

        }, {
            key: "route",
            value: function route() {
                if (!settingURL) {
                    var matchFound = matchAndCall();
                    if (!matchFound) {
                        throw new Error('No route registered for this url : ' + getPathName());
                    }
                } else {
                    settingURL = false;
                }
                loop(listeners, function (listener_name, listener_fn) {
                    call_fn(listener_fn);
                });
            }

            /**
             * @function
             * @static
             * @returns {RouterURLConf} Current view name.
             * */

        }, {
            key: "getCurrentRoute",
            value: function getCurrentRoute() {
                if (!currentRoute) {
                    currentRoute = matchUrl().opts;
                }
                return currentRoute;
            }

            /**
             * @function
             * @static
             * @returns {String} Current view name.
             * */

        }, {
            key: "getCurrentPageName",
            value: function getCurrentPageName() {
                return AdharaRouter.getCurrentRoute().view_name;
            }

            /**
             * @function
             * @static
             * @returns {String} Current URL.
             * */

        }, {
            key: "getCurrentURL",
            value: function getCurrentURL() {
                return getFullPath();
            }

            /**
             * @function
             * @static
             * @returns {String} Current Path.
             * */

        }, {
            key: "getCurrentPath",
            value: function getCurrentPath() {
                return getPathName();
            }

            /**
             * @function
             * @static
             * @returns {String} Current URL's Hash.
             * */

        }, {
            key: "getCurrentHash",
            value: function getCurrentHash() {
                return getHash();
            }

            /**
             * @function
             * @private
             * @param {URL|String} url - URL to be navigated to.
             * @description
             * navigates to the provided URL.
             * */

        }, {
            key: "go",
            value: function go(url) {
                if (isCurrentPath(url)) {
                    return false;
                }
                history.pushState(_defineProperty({}, STATE_KEY, true), parent.document.title, url);
                //considering the behaviour of immediate state change
                var state = history.state;
                updateHistoryStack();
                pathParams = {};
                if (state !== undefined && state !== '') {
                    // let data = state.data;
                    if (state[STATE_KEY] === true || state.data[STATE_KEY] === true) {
                        this.route();
                    }
                }
                // History.pushState({[STATE_KEY]:true}, parent.document.title, url);
                return true;
            }

            /**
             * @function
             * @static
             * @param {String} url - URL that to be updated.
             * @description Updates current URL in the path but doesn't call the view function. Appends current URL in History.
             * */

        }, {
            key: "setURL",
            value: function setURL(url) {
                url = this.transformURL(url);
                settingURL = true;
                var gone = this.go(url);
                if (!gone) {
                    settingURL = false;
                }
            }

            /**
             * @function
             * @static
             * @param {!String} url - URL that to be updated.
             * @description Replaces current URL in the path, and calls view function. No modifications made to History.
             * */

        }, {
            key: "overrideURL",
            value: function overrideURL(url) {
                url = this.transformURL(url);
                history.replaceState(_defineProperty({}, STATE_KEY, true), parent.document.title, url);
                // History.replaceState({[STATE_KEY]:true}, parent.document.title, url);
            }

            /**
             * @function
             * @static
             * @param {!String} url - URL that to be updated.
             * @description Replaces current URL in the path but doesn't call the view function. No modifications made to History.
             * */

        }, {
            key: "updateURL",
            value: function updateURL(url) {
                settingURL = true;
                this.overrideURL(url);
            }

            /**
             * @function
             * @static
             * @param {!String} url - URL to be navigated to.
             * @param {Boolean} [force=false] - If true, will call the view function even if current URL is same as provided URL.
             * @description Navigates to the provided URL.
             * And if provided URL is same as current URL, view function will not be called unless force parameter is true.
             * */

        }, {
            key: "navigateTo",
            value: function navigateTo(url, force) {
                url = this.transformURL(url);
                if (!isCurrentPath(url)) {
                    var gone = this.go(url);
                    if (gone) {
                        fetchQueryParams();
                    }
                } else if (force) {
                    this.route();
                }
            }

            /**
             * @function
             * @static
             * @param {!String} backwardURL - backward URL to be navigated to.
             * @description Navigates to the backward URL by moving back in history.
             * In case if history is not available, Router will navigate to the backwardURL as if it is a new URL.
             * */

        }, {
            key: "goBack",
            value: function goBack(backwardURL) {
                if (!backwardURL) {
                    backwardURL = historyStack.slice(-1)[0];
                    if (!backwardURL) {
                        return false;
                    }
                }
                backwardURL = this.transformURL(backwardURL);
                var backwardIndex = historyStack.lastIndexOf(backwardURL);
                if (backwardIndex === -1) {
                    this.navigateTo(backwardURL);
                } else {
                    var stackLen = historyStack.length;
                    var negativeIndex = backwardIndex - stackLen;
                    historyStack.splice(stackLen + negativeIndex, -negativeIndex);
                    history.go(negativeIndex);
                }
            }

            /**
             * @function
             * @static
             * @returns {Array<String>} List of visited URL's that are routed by Router.
             * */

        }, {
            key: "peekStack",
            value: function peekStack() {
                return historyStack.slice();
            }

            /**
             * @function
             * @static
             * @param {String} param_name - name of the query parameter for which value is required.
             * @returns {String} Value of the search parameter from current URL.
             * */

        }, {
            key: "getQueryParam",
            value: function getQueryParam(param_name) {
                updateParams();
                return queryParams[param_name];
            }
        }, {
            key: "getPathParam",
            value: function getPathParam(param_name) {
                return pathParams[param_name];
            }

            /**
             * Route type
             * @typedef {String} RouteType
             * @see {@link Router.RouteTypes}
             * */

            /**
             * @function
             * @static
             * @param {!Object} new_params - Search Parameters to be added/updated.
             * @param {Boolean} [drop_existing=false] - Whether to completely replace existing search parameters or update with the new ones.
             * @param {RouteType} [route_type=this.routeTypes.NAVIGATE] - How to route. {@link RouteType}
             * @param {Object} route_options - Route options.
             * @param {boolean} [route_options.force=false] - Whether to force navigate URL or not. Will call view function even if current URL is same. Applies for {@link RouteType.NAVIGATE}.
             * @description
             * Updates the query parameters and navigates to the new URL based on route_type that is constructed with the new parameters.
             * */

        }, {
            key: "updateQueryParams",
            value: function updateQueryParams(new_params, drop_existing, route_type, route_options) {
                if (!route_type) {
                    route_type = this.RouteTypes.NAVIGATE;
                }
                if (!Object.keys(new_params).length) {
                    return;
                }
                if (drop_existing) {
                    queryParams = new_params;
                } else {
                    Object.assign(queryParams, new_params);
                }
                _route(route_type, route_options);
            }

            /**
             * @function
             * @static
             * @param {!Array} param_keys - Search Parameters to be dropped.
             * @param {RouteType} [route_type=this.routeTypes.NAVIGATE] - How to route. {@link RouteType}
             * @param {Object} route_options - Route options.
             * @param {boolean} [route_options.force=false] - Whether to force navigate URL or not. Will call view function even if current URL is same. Applies for {@link RouteType.NAVIGATE}.
             * @description
             * Drops provided keys if exist, updates the query parameters and navigates to the new URL that is constructed with the new parameters.
             * */

        }, {
            key: "removeQueryParams",
            value: function removeQueryParams(param_keys, route_type, route_options) {
                loop(param_keys, function (idx, key) {
                    delete queryParams[key];
                });
                _route(route_type, route_options);
            }

            /**
             * @function
             * @static
             * @param {String} view_name - View name with which a URL pattern is registered.
             * @returns {RegExp} Pattern for which the provided view name is the view name.
             * */

        }, {
            key: "getURLPatternByPageName",
            value: function getURLPatternByPageName(view_name) {
                var matched_pattern = null;
                loop(registeredUrlPatterns, function (pattern, options) {
                    if (options.view_name === view_name) {
                        matched_pattern = pattern;
                        return false;
                    }
                });
                return matched_pattern;
            }
        }]);

        return Router;
    }();

    //---------------------

    AdharaRouter = Router;

    /**
     * @description
     * setting current view name to undefined on moving to some other page that is not registered with router.
     * */
    window.onpopstate = function (e) {
        // if(e.state[STATE_KEY]){
        AdharaRouter.route();
        // }else{
        //     currentRoute = null;
        // }
        /*if(currentRoute){
            let url_pattern = Router.getURLPatternByPageName(currentRoute.view_name);
            if(!(url_pattern && new RegExp(url_pattern).test(getPathName()))){
                currentRoute = undefined;
            }
        }*/
    };

    /**
     * @static
     * @readonly
     * @enum {RouteType}
     * @description Operations that can be performed on the entity
     * */
    AdharaRouter.RouteTypes = Object.freeze({
        /**
         * navigate to the url and call the view function. This will add current URL to History.
         * */
        NAVIGATE: "navigate",
        /**
         * override the current url and call the view function. This will not change History.
         * */
        OVERRIDE: "override",
        /**
         * Just update the URL. Adds current URL to History. View function will not be called.
         * */
        SET: "set",
        /**
         * Overrides the current URL. No modifications made to History. View function will not be called.
         * */
        UPDATE: "update"
    });

    AdharaRouter.enableAllAnchors = true;

    AdharaRouter.listen = function () {
        /**
         * Listening to elements with route property in DOM.
         * If it has href attribute, preventing default event and proceeding with SdpRouting
         * */

        function getRoutingElement(event) {
            return event.target.nodeName !== "A" ? event.currentTarget : event.target;
        }

        function hasAttribute(elem, attribute_name) {
            return elem.hasAttribute(attribute_name);
        }

        function routeHandler(event) {
            var re = getRoutingElement(event);
            if ((AdharaRouter.enableAllAnchors || hasAttribute(re, "route")) && hasAttribute(re, "href")) {
                var url = this.getAttribute('href').trim();
                var target = this.getAttribute("target");
                if (target && target !== "_self" || url.indexOf('javascript') !== -1) {
                    return;
                }
                try {
                    var go_to_url = new URL(url);
                    if (go_to_url.host !== window.location.host) {
                        return;
                    }
                } catch (e) {/*Do nothing. Try catches the invalid url on new ULR(...)*/}
                if (url) {
                    event.preventDefault();
                    event.stopPropagation();
                }
                var go_back = this.getAttribute("data-back");
                var force = this.getAttribute("data-force") !== "false";
                if (go_back) {
                    return AdharaRouter.goBack(url);
                }
                AdharaRouter.navigateTo(url, force);
            }
        }

        jQuery(document).on("click", "a", routeHandler);
        jQuery(document).on("click", "[route]", routeHandler);

        /*document.addEventListener('click', function (e) {
            if(e.target.nodeName === "A" && ( AdharaRouter.enableAllAnchors || hasAttribute(e.target, "route") ) ){
                let url = e.target.getAttribute('href').trim();
                if(url.indexOf('javascript') !== -1){return;}
                if(url){
                    e.preventDefault();
                    e.stopPropagation();
                }
                let go_back = e.target.getAttribute("data-back");
                let force = e.target.getAttribute("data-force") !== "false";
                if(go_back){ return AdharaRouter.goBack(url); }
                AdharaRouter.navigateTo(url, force);
            }
          }, false);*/
    };

    //---------------------


    updateHistoryStack(); //updating history stack with the entry URL
})();
/**------------------------------------------------------------------------------------------------------------------**/
/**
 * Created by rohit on 25/5/16.
 */

var RestAPI = {};

(function () {

    RestAPI.error_codes = {
        not_found: 404
    };

    RestAPI.getCookie = function (name) {
        var cookieValue = null;
        if (document.cookie && document.cookie !== '') {
            var cookies = document.cookie.split(';');
            for (var i = 0; i < cookies.length; i++) {
                var cookie = jQuery.trim(cookies[i]);
                // Does this cookie string begin with the name we want?
                if (cookie.substring(0, name.length + 1) === name + '=') {
                    cookieValue = decodeURIComponent(cookie.substring(name.length + 1));
                    break;
                }
            }
        }
        return cookieValue;
    };

    RestAPI.getHeaders = function (url, method) {
        if (isFullURL(url)) {
            return {};
        }
        return {
            'X-CSRFToken': RestAPI.getCookie('csrftoken')
        };
    };

    RestAPI.handle_api_success = function (fns, fne, d, s, x, toastFailure, successMessage) {
        //data, success, xhr
        if (s === "nocontent") {
            //DO NOTHING
            return;
        }
        if (s === "success") {
            if (x.getResponseHeader('Content-Disposition')) {
                return call_fn(fns, d, x);
            } else {
                if (typeof d === "string") {
                    d = JSON.parse(d);
                }
                if (d.status === "success") {
                    if (successMessage && successMessage !== "") {
                        Toast.success(successMessage);
                    }
                    return call_fn(fns, d, x);
                }
            }
        }
        if (toastFailure !== false) {
            Toast.error(d, x);
        }
        call_fn(fne, d, x);
    };

    RestAPI.handle_api_failure = function (fne, x, s, e) {
        //xhr, status, error
        if (x.readyState === 0 && x.status === 0 && e === "") {
            e = "Unable to connect to server";
        }
        call_fn(fne, e, x);
        // if(x.responseText){
        //     return Toast.error(e+"\n"+x.responseText);
        // }
        // Toast.error(e);
    };

    RestAPI.post = function (o) {
        o.type = "post";
        send(o);
    };

    RestAPI.get = function (o) {
        o.type = "get";
        send(o);
    };

    RestAPI.put = function (o) {
        o.type = "put";
        send(o);
    };

    RestAPI.patch = function (o) {
        o.type = "patch";
        send(o);
    };

    RestAPI.delete = function (o) {
        o.type = "delete";
        send(o);
    };

    function isFullURL(link) {
        return link.startsWith("http://") || link.startsWith("https://") || link.startsWith("//");
    }

    function formatURL(url) {
        if (isFullURL(url)) {
            return url;
        }
        var base = Adhara.app ? Adhara.app.apiServerURL : "/";
        if (base.lastIndexOf("/") + 1 !== base.length) {
            base += "/";
        }
        if (url.indexOf("/") === 0) {
            url = url.substring(1);
        }
        return base + url;
    }

    function send(o) {
        o.url = formatURL(o.url);
        var fns = o.success;
        var fnf = o.failure;
        o.success = function (d, s, x) {
            RestAPI.handle_api_success(fns, fnf, d, s, x, o.handleError, o.successMessage);
        };
        o.error = function (x, s, e) {
            RestAPI.handle_api_failure(fnf, x, s, e);
        };
        o.headers = RestAPI.getHeaders(o.url, o.type);
        if (o.data instanceof FormData) {
            multipart(o);
        } else {
            ajax(o);
        }
    }

    function ajax(o) {
        if (o.type !== "get" && o.data instanceof Object) {
            if (!Object.keys(o.data).length) {
                delete o.data;
            } else {
                o.data = JSON.stringify(o.data);
            }
        }
        /*if(o.type === "get" || o.type === "patch" || o.type === "post") {
            if (o.type === "get") {
                /!*if (o.data) {
                    o.data = {data: o.data};
                }*!/
            }
        }*/
        o.headers['Content-Type'] = 'application/json';
        jQuery.ajax(o);
    }

    function multipart(o) {
        var xhr = new XMLHttpRequest();
        xhr.open(o.type.toUpperCase(), o.url);
        loop(o.headers, function (header, value) {
            xhr.setRequestHeader(header, value);
        });
        xhr.onreadystatechange = function () {
            if (xhr.readyState === XMLHttpRequest.DONE) {
                if (xhr.status === 200) {
                    o.success(xhr.responseText, "success", xhr);
                } else {
                    o.error(xhr, "error", "Multipart upload error");
                }
            }
        };
        xhr.send(o.data);
    }
})();
/**------------------------------------------------------------------------------------------------------------------**/
var Toast = {};
(function () {

    Toast.make = function (title, content, type) {
        //type should be "success", "error" or "info"
        if (Adhara.app) {
            return Adhara.app.toast(title, content, type);
        } else {
            if (window.ToastHandler) {
                return window.ToastHandler[type](content, title);
            } else {
                console.log("No toast handler found. Toast message: type-" + type + ", title-" + title + ", content=" + content);
            }
        }
    };

    Toast.error = function (message) {
        return Toast.make("Error!", message, "error");
    };

    Toast.success = function (message) {
        return Toast.make("Success", message, "success");
    };
})();

var AdharaDefaultToaster = {};

(function () {

    var notifyQueue = 0;

    function notify(title, content, type, id) {
        id = "notification_" + id;
        var $notificationDiv = jQuery('<div class="notification ' + type + '" id="' + id + '"/>');
        var $title = jQuery('<strong style="display:block" />').text(title);
        var $content = jQuery('<p />').html(content);
        $notificationDiv.append($title).append($content);
        jQuery('body').append($notificationDiv);

        var existing_notifications = jQuery('.notification');
        if (existing_notifications.length > 0) {
            jQuery.each(existing_notifications, function (index, value) {
                if (jQuery(this).attr('id') !== id) {
                    var new_top = parseFloat(jQuery(this).css('top')) + $notificationDiv.height() + 40;
                    jQuery(this).animate({ top: new_top + "px" }, 'fast');
                }
            });
        }

        jQuery('#' + id).animate({ right: '0px' }, 'fast', function () {
            notifyQueue--;
        });
        setTimeout(function () {
            jQuery('#' + id).animate({ opacity: '0.5' }, 'fast', function () {
                jQuery('#' + id).hide('slide', { direction: 'right' }, 'slow', function () {
                    setTimeout(function () {
                        jQuery('#' + id).remove();
                    }, 1000);
                });
            });
        }, 5000);
    }

    AdharaDefaultToaster.make = function (title, content, type) {
        var maxNotificationsHeight = window.innerHeight - 200;
        var currentNotificationsHeight = 0;
        jQuery.each(jQuery('.notification'), function () {
            currentNotificationsHeight += jQuery(this).height() + 40;
        });
        var proceed = currentNotificationsHeight < maxNotificationsHeight;
        var notifyTimeout = notifyQueue === 0 ? 0 : 2000;
        if (!content && type === "failure") {
            content = "Error occurred...";
        }
        var nid = performance.now();
        setTimeout(function () {
            if (notifyQueue === 0 && proceed) {
                notifyQueue++;
                notify(title, content, type, nid);
            } else {
                AdharaDefaultToaster.make(title, content, type);
            }
        }, notifyTimeout);
        return {
            remove: function remove() {
                jQuery("#notification_" + nid).remove();
            }
        };
    };
})();
/**------------------------------------------------------------------------------------------------------------------**/
function deserialize(str, context) {
    if ((typeof context === "undefined" ? "undefined" : _typeof(context)) === 'object' && context.blob) {
        var data = JSON.parse(str);
        return new context.blob(data);
    }
    if (context) {
        var _data = JSON.parse(str);
        return new context.blob(_data);
    }
    throw new Error("context passed is invalid: " + context);
}

function isPrimitive(ob) {
    return ob !== Object(ob);
}

var Serializable = function () {
    _createClass(Serializable, null, [{
        key: "_validateData",
        value: function _validateData(data) {
            // checks if there is information loss on serializing the data
            function equivalence(ob1, ob2) {
                if (isPrimitive(ob1) && isPrimitive(ob2)) {
                    return ob1 === ob2;
                } else if (isPrimitive(ob1) !== isPrimitive(ob2)) {
                    return false;
                }
                for (var key in ob1) {
                    if (ob1.hasOwnProperty(key)) {
                        if (!equivalence(ob1[key], ob2[key])) {
                            return false;
                        }
                    }
                }
                return true;
            }
            if (isPrimitive(data)) {
                return true;
            } else if ((typeof data === "undefined" ? "undefined" : _typeof(data)) !== 'object' || data === null || data instanceof Array) {
                // console.warn("null/array data", data);
                return true; //Todo handle array of arrays as well !
            } else {
                var eq_success = equivalence(JSON.parse(JSON.stringify(data)), data);
                if (!eq_success) {
                    console.warn("Equivalence check failed for the data. Data is not serializable.", data);
                }
                return eq_success;
            }
        }
    }]);

    function Serializable(data) {
        _classCallCheck(this, Serializable);

        if (!Serializable._validateData(data)) {
            throw new Error("blob must be initialized with serializable(tree-like, non-array) object: " + data);
        }
    }

    _createClass(Serializable, [{
        key: "serialize",
        value: function serialize() {
            throw new Error("Need to define method: serialize");
        }
    }]);

    return Serializable;
}();
/**------------------------------------------------------------------------------------------------------------------**/


var DataBlob = function (_Serializable) {
    _inherits(DataBlob, _Serializable);

    // extender blobs represent tree like objects only

    function DataBlob(data) {
        _classCallCheck(this, DataBlob);

        var _this3 = _possibleConstructorReturn(this, (DataBlob.__proto__ || Object.getPrototypeOf(DataBlob)).call(this, data));

        if (!data) {
            return _possibleConstructorReturn(_this3);
        }
        _this3._data = data;
        var data_is_fine = _this3.validate(data);
        if (data_is_fine === false) {
            throw new Error("assigned data does not clear the custom validation: " + data);
        }
        _this3.onInit();
        return _this3;
    }

    /**
     * Easy override method: to handle creating any cacheable objects/references classes
     * */


    _createClass(DataBlob, [{
        key: "onInit",
        value: function onInit() {}
    }, {
        key: "validate",
        value: function validate() {
            return true;
        }
    }, {
        key: "_formatTimeBit",
        value: function _formatTimeBit(value) {
            return (value.toString().length === 1 ? "0" : "") + value.toString();
        }

        //TODO generalize this using string format's

    }, {
        key: "_formatTime",
        value: function _formatTime(date) {
            return this._formatTimeBit(date.getHours()) + ":" + this._formatTimeBit(date.getMinutes()) + ":" + this._formatTimeBit(date.getSeconds());
        }

        //TODO generalize this using string format's

    }, {
        key: "_formatDate",
        value: function _formatDate(date) {
            return this._formatTimeBit(date.getDate()) + "/" + this._formatTimeBit(date.getMonth()) + "/" + date.getFullYear();
        }
    }, {
        key: "data",


        // noinspection JSAnnotator
        get: function get() {
            return this._data;
        },
        set: function set(data) {
            if (!Serializable._validateData(data)) {
                // all blobs are validated for serializability
                throw new Error("blob must represent a non-array, tree-like object " + data);
            }
            var data_is_fine = this.validate(data); // custom validators must throw their own errors or return a false status on failure
            if (data_is_fine === false) {
                throw new Error("assigned data does not clear the custom validation: " + data);
            }
            this._data = data;
        }
    }, {
        key: "packet",
        get: function get() {
            return this.serialize();
        }
    }], [{
        key: "_context",
        get: function get() {
            return "base";
        }
    }]);

    return DataBlob;
}(Serializable);
/**------------------------------------------------------------------------------------------------------------------**/


var AdharaController = function () {
    function AdharaController() {
        _classCallCheck(this, AdharaController);
    }

    _createClass(AdharaController, [{
        key: "control",


        /**
         * @function
         * @instance
         * @param {String} method - can be any of [get|get_list|put|post|delete]
         * @param {Object} entity_config - entity configuration
         * @param {*} [data=undefined] - data to be passed as a payload/query params
         * @param {Object} [options={}] - options for DI handling! might not be necessary
         * @description initial hook to handle a data call
         * */
        value: function control(method, entity_config, data, options) {
            return Adhara.dataInterface.enqueue(method, entity_config, data, options);
        }

        /**
         * @function
         * @instance
         * @param {Object} entity_config - entity configuration
         * @param {*} [data=undefined] - data to be passed as a payload/query params
         * @param {Object} [options={}] - options for DI handling! might not be necessary
         * @description initial hook to handle a get data call
         * */

    }, {
        key: "getData",
        value: function getData(entity_config, data, options) {
            this.control('get', entity_config, data, options);
        }
    }, {
        key: "getListData",


        /**
         * @function
         * @instance
         * @param {Object} entity_config - entity configuration
         * @param {*} [data=undefined] - data to be passed as a payload/query params
         * @param {Object} [options={}] - options for DI handling! might not be necessary
         * @description initial hook to handle a get list data call
         * */
        value: function getListData(entity_config, data, options) {
            this.control('get_list', entity_config, data, options);
        }
    }, {
        key: "putData",


        /**
         * @function
         * @instance
         * @param {Object} entity_config - entity configuration
         * @param {*} [data=undefined] - data to be passed as a payload/query params
         * @param {Object} [options={}] - options for DI handling! might not be necessary
         * @description initial hook to handle a put data call
         * */
        value: function putData(entity_config, data, options) {
            this.control('put', entity_config, data, options);
        }
    }, {
        key: "postData",


        /**
         * @function
         * @instance
         * @param {Object} entity_config - entity configuration
         * @param {*} [data=undefined] - data to be passed as a payload/query params
         * @param {Object} [options={}] - options for DI handling! might not be necessary
         * @description initial hook to handle a post data call
         * */
        value: function postData(entity_config, data, options) {
            this.control('post', entity_config, data, options);
        }
    }, {
        key: "deleteData",


        /**
         * @function
         * @instance
         * @param {Object} entity_config - entity configuration
         * @param {*} [data=undefined] - data to be passed as a payload/query params
         * @param {Object} [options={}] - options for DI handling! might not be necessary
         * @description initial hook to handle a delete data call
         * */
        value: function deleteData(entity_config, data, options) {
            this.control('delete', entity_config, data, options);
        }
    }]);

    return AdharaController;
}();
/**------------------------------------------------------------------------------------------------------------------**/
/**
 * @class
 * @classdesc a base class that is to be extended by all the view classes
 * */


var AdharaView = function (_AdharaController) {
    _inherits(AdharaView, _AdharaController);

    /**
     * @constructor
     * @param {AdharaView} parentViewInstance - parent view instance that is to be passed to render a view inside another.
     * */
    function AdharaView(parentViewInstance) {
        _classCallCheck(this, AdharaView);

        var _this4 = _possibleConstructorReturn(this, (AdharaView.__proto__ || Object.getPrototypeOf(AdharaView)).call(this));

        _this4._parentViewInstance = parentViewInstance;
        Adhara.addViewToInstances(_this4);
        _this4._data = null;
        _this4._state = {};
        _this4._error = null;
        _this4._event_listeners = {};
        _this4.is_active = false;
        _this4._registerEvents(["ViewRendered", "SubViewsRendered", "ViewFormatted", "ViewDestroyed"]);
        _this4._registerEvents(_this4.events);
        _this4.onInit();
        return _this4;
    }

    /**
     * @function
     * @instance
     * @description Anything that needs to be done right after initializing the view.
     * */


    _createClass(AdharaView, [{
        key: "onInit",
        value: function onInit() {}
        //Override this method to do any miscellaneous operations/assignments right after initializing the View


        /**
         * @getter
         * @instance
         * @returns {Array<String>} return list of custom event names in PascalCase...
         * @description event listener registration functions will be created dynamically by prepending "on" th the event name
         * @example
         * say get events() returns this array: ["Render", "Format"]
         * listener functions these events can be then registered using `onRender(listener)` and `onFormat(listener)`
         * */

    }, {
        key: "_registerEvents",


        /**
         * @function
         * @protected
         * @param {Array<String>} event_names - list of events that are to be enabled on the current View instance.
         * @description registers the event names and creates event registration functions.
         * @example
         * _registerEvents(["Render", "Format"]);
         * //This will create 2 functions onRender and onFormat runtime to allow registration of events
         * */
        value: function _registerEvents(event_names) {
            var _this5 = this;

            var _loop = function _loop(event_name) {
                _this5._event_listeners[event_name] = [];
                _this5["on" + event_name] = function (handler) {
                    _this5._event_listeners[event_name].push(handler);
                };
                _this5["off" + event_name] = function (handler) {
                    _this5._event_listeners[event_name].splice(_this5._event_listeners[event_name].indexOf(handler), 1);
                };
            };

            var _iteratorNormalCompletion4 = true;
            var _didIteratorError4 = false;
            var _iteratorError4 = undefined;

            try {
                for (var _iterator4 = event_names[Symbol.iterator](), _step4; !(_iteratorNormalCompletion4 = (_step4 = _iterator4.next()).done); _iteratorNormalCompletion4 = true) {
                    var event_name = _step4.value;

                    _loop(event_name);
                }
            } catch (err) {
                _didIteratorError4 = true;
                _iteratorError4 = err;
            } finally {
                try {
                    if (!_iteratorNormalCompletion4 && _iterator4.return) {
                        _iterator4.return();
                    }
                } finally {
                    if (_didIteratorError4) {
                        throw _iteratorError4;
                    }
                }
            }
        }
    }, {
        key: "trigger",
        value: function trigger(event_name) {
            for (var _len3 = arguments.length, data = Array(_len3 > 1 ? _len3 - 1 : 0), _key3 = 1; _key3 < _len3; _key3++) {
                data[_key3 - 1] = arguments[_key3];
            }

            var _iteratorNormalCompletion5 = true;
            var _didIteratorError5 = false;
            var _iteratorError5 = undefined;

            try {
                for (var _iterator5 = this._event_listeners[event_name][Symbol.iterator](), _step5; !(_iteratorNormalCompletion5 = (_step5 = _iterator5.next()).done); _iteratorNormalCompletion5 = true) {
                    var event_handler = _step5.value;

                    event_handler(data);
                }
            } catch (err) {
                _didIteratorError5 = true;
                _iteratorError5 = err;
            } finally {
                try {
                    if (!_iteratorNormalCompletion5 && _iterator5.return) {
                        _iterator5.return();
                    }
                } finally {
                    if (_didIteratorError5) {
                        throw _iteratorError5;
                    }
                }
            }
        }

        /**
         * @method
         * @getter
         * @returns {HandlebarTemplate} template of the view
         * */

    }, {
        key: "getCustomTemplate",
        value: function getCustomTemplate(type) {
            var _iteratorNormalCompletion6 = true;
            var _didIteratorError6 = false;
            var _iteratorError6 = undefined;

            try {
                for (var _iterator6 = Adhara.viewHierarchy[Symbol.iterator](), _step6; !(_iteratorNormalCompletion6 = (_step6 = _iterator6.next()).done); _iteratorNormalCompletion6 = true) {
                    var ViewClass = _step6.value;

                    if (ViewClass.isPrototypeOf(this.constructor)) {
                        if (Adhara.app.customViewConfig[type] && Adhara.app.customViewConfig[type][ViewClass.name]) {
                            return Adhara.app.customViewConfig[type][ViewClass.name];
                        }
                    }
                }
            } catch (err) {
                _didIteratorError6 = true;
                _iteratorError6 = err;
            } finally {
                try {
                    if (!_iteratorNormalCompletion6 && _iterator6.return) {
                        _iterator6.return();
                    }
                } finally {
                    if (_didIteratorError6) {
                        throw _iteratorError6;
                    }
                }
            }
        }

        /**
         * @method
         * @getter
         * @returns {HandlebarTemplate} template of the view
         * */

    }, {
        key: "getTemplate",


        /**
         * @function
         * @instance
         * @description Helper method to get required template. Error template or success template.
         * */
        value: function getTemplate() {
            return this.state.fetching_data ? this.fetchingDataTemplate : this.errors ? this.errorTemplate || this.template : this.template;
        }

        /**
         * @getter
         * @instance
         * @param {String} [batch_identifier=undefined] - In case if the config is a batch config,
         * use batch_identifier to know what url's path parameters are required.
         * @returns {Object|null} url path parameters as an object with keys as template variables
         * */

    }, {
        key: "getURLPathParams",
        value: function getURLPathParams(batch_identifier) {
            return null;
        }

        /**
         * @getter
         * @instance
         * @returns {Object|null} Add dynamic input data to API calls made to server
         * */

    }, {
        key: "formatURL",
        value: function formatURL(url, params) {
            if (!params) {
                return url;
            }
            var match = url.match(/\${([a-zA-Z0-9$_]*)}/gi);
            var this_match = match.map(function (match) {
                return '${this.' + /\${([a-zA-Z0-9$_]*)}/.exec(match)[1] + '}';
            });
            for (var idx in match) {
                //in works of doesn't work here as it is not a classic Array
                url = url.replace(match[idx], this_match[idx]);
            }
            return new Function("return `" + url + "`;").call(params);
        }
    }, {
        key: "formatEntityConfig",
        value: function formatEntityConfig(entity_config) {
            if (entity_config.data_config.hasOwnProperty("batch_data_override")) {
                var _iteratorNormalCompletion7 = true;
                var _didIteratorError7 = false;
                var _iteratorError7 = undefined;

                try {
                    for (var _iterator7 = entity_config.data_config.batch_data_override[Symbol.iterator](), _step7; !(_iteratorNormalCompletion7 = (_step7 = _iterator7.next()).done); _iteratorNormalCompletion7 = true) {
                        var one_config = _step7.value;

                        var url_path_params = this.getURLPathParams(one_config.identifier);
                        if (url_path_params) {
                            one_config.url = this.formatURL(one_config.url, url_path_params);
                        }
                    }
                } catch (err) {
                    _didIteratorError7 = true;
                    _iteratorError7 = err;
                } finally {
                    try {
                        if (!_iteratorNormalCompletion7 && _iterator7.return) {
                            _iterator7.return();
                        }
                    } finally {
                        if (_didIteratorError7) {
                            throw _iteratorError7;
                        }
                    }
                }
            } else {
                var _url_path_params = this.getURLPathParams();
                if (_url_path_params) {
                    entity_config.data_config.url = this.formatURL(entity_config.data_config.url, _url_path_params);
                }
            }
            return entity_config;
        }

        /**
         * @function
         * @instance
         * @returns Adhara style entity config for entity mapped with this view
         * */

    }, {
        key: "getPayload",


        /**
         * @function
         * @instance
         * @returns Add dynamic input data to API calls made to server
         * @description intermediate function that can be overridden by other generic views
         * */
        value: function getPayload() {
            return this.payload;
        }

        /**
         * @getter
         * @instance
         * @returns {Object} state object of the current View
         * */

    }, {
        key: "create",
        value: function create(parentViewInstance) {
            if (parentViewInstance && AdharaView.isPrototypeOf(parentViewInstance.constructor)) {
                this._parentViewInstance = parentViewInstance;
            }
            Adhara.addToActiveViews(this);
            this.is_active = true;
            this.fetchData();
        }
    }, {
        key: "isActive",
        value: function isActive() {
            return Adhara.isActiveView(this) && this.is_active;
        }

        /**
         * @function
         * @instance
         * @description hook to make API calls, data change event will be triggered on successful API call.
         * By default no API call will be made and dataChange method will be called right away.
         * */

    }, {
        key: "fetchData",
        value: function fetchData() {
            this._state.fetching_data = true;
            var config = this.entityConfig;
            if (config) {
                if (config.data_config.hasOwnProperty("batch_data_override")) {
                    this.render();
                    return this.control(undefined, config);
                }
                if (Adhara.dataInterface.getHTTPMethod(config.data_config.default_query_type) === "get") {
                    this.render();
                    return this.control(config.data_config.default_query_type, config, this.getPayload());
                }
                return this.handleDataChange();
            }
            this.handleDataChange();
        }

        /**
         * @function
         * @protected
         * @param {DataBlob|Array<DataBlob>} [new_data=null] - Data in the form of DataBlob instance to be updated in view
         * */

    }, {
        key: "handleDataChange",
        value: function handleDataChange(new_data) {
            if (!this.isActive()) {
                return;
            }
            this.dataChange(new_data);
            this._state.fetching_data = false;
            this.render();
        }

        /**
         * @function
         * @private
         * @param {*} error - Error to be updated in view
         * */

    }, {
        key: "handleDataError",
        value: function handleDataError(error) {
            if (!this.isActive()) {
                return;
            }
            this.dataError(error);
            this._state.fetching_data = false;
            this.render();
        }

        /**
         * @function
         * @private
         * @param {Object<String, Object<String, DataBlob|*>>} map - Data/Error to be updated in view. Map of {String} identifier vs {Object<success:response|error:error>}
         * */

    }, {
        key: "handleBatchData",
        value: function handleBatchData(map) {
            if (!this.isActive()) {
                return;
            }
            var errors = {};
            var success = {};
            for (var identifier in map) {
                if (map.hasOwnProperty(identifier)) {
                    if (map[identifier].hasOwnProperty("error")) {
                        errors[identifier] = map[identifier].error;
                    } else {
                        success[identifier] = map[identifier].success;
                    }
                }
            }
            if (Object.keys(success).length) {
                this.dataChange(success);
            }
            if (Object.keys(errors).length) {
                this.dataError(errors);
            }
            this._state.fetching_data = false;
            this.render();
        }

        /**
         * @function
         * @instance
         * @param {DataBlob|*} new_data - Data in the form of DataBlob instance to be set in view
         * */

    }, {
        key: "dataChange",
        value: function dataChange(new_data) {
            this._data = new_data;
            this._error = null;
        }

        /**
         * @function
         * @instance
         * @param {*} error - Error to be set in view
         * */

    }, {
        key: "dataError",
        value: function dataError(error) {
            this._error = error;
        }

        /**
         * @method
         * @getter
         * @returns {*} View errors, that can be consumed by the template.
         * */

    }, {
        key: "_getHTML",
        value: function _getHTML(template) {
            return HandlebarUtils.execute(template || this.getTemplate(), this);
        }
    }, {
        key: "_getParentContainer",
        value: function _getParentContainer() {
            var container = this.parentView.contentContainer;
            if (typeof container === "string") {
                return container;
            }
            return container[this.constructor.name] || container["*"];
        }
    }, {
        key: "getParentContainerElement",
        value: function getParentContainerElement() {
            return document.querySelector(this._getParentContainer());
        }
    }, {
        key: "_render",
        value: function _render() {}
    }, {
        key: "render",
        value: function render() {
            var _this6 = this;

            var container = this.getParentContainerElement();
            if (!container) {
                console.warn("No container defined/available", this.constructor.name);
                return;
            }
            container.innerHTML = this._getHTML();
            if (this.state.fetching_data) {
                return;
            }
            this.trigger("ViewRendered");
            this._format(container);
            setTimeout(function () {
                _this6.format(container);
                _this6.trigger("ViewFormatted");
            }, 0);
            this.renderSubViews();
            this.trigger("SubViewsRendered");
        }

        /**
         * @function
         * @getter
         * @returns {String} A CSS selector inside which child views are to be rendered.
         * */

    }, {
        key: "renderSubViews",
        value: function renderSubViews() {
            if (!this.subViews) {
                return;
            }
            var _iteratorNormalCompletion8 = true;
            var _didIteratorError8 = false;
            var _iteratorError8 = undefined;

            try {
                for (var _iterator8 = this.subViews[Symbol.iterator](), _step8; !(_iteratorNormalCompletion8 = (_step8 = _iterator8.next()).done); _iteratorNormalCompletion8 = true) {
                    var sub_view = _step8.value;

                    Adhara.createView(Adhara.getView(sub_view, this), this);
                }
            } catch (err) {
                _didIteratorError8 = true;
                _iteratorError8 = err;
            } finally {
                try {
                    if (!_iteratorNormalCompletion8 && _iterator8.return) {
                        _iterator8.return();
                    }
                } finally {
                    if (_didIteratorError8) {
                        throw _iteratorError8;
                    }
                }
            }
        }
    }, {
        key: "_format",
        value: function _format(container) {
            var _this7 = this;

            var _loop2 = function _loop2(action) {
                var onActionElements = container.querySelectorAll("[data-on" + action + "]");

                var _loop3 = function _loop3(actionElement) {
                    if (actionElement.dataset['_adharaevent_']) {
                        return "continue";
                    }
                    actionElement.addEventListener(action, function (event) {
                        var data = actionElement.dataset;
                        var action_key = "on" + action;
                        if (_this7[data[action_key]]) {
                            _this7[data[action_key]](event, data, _this7);
                        } else {
                            var fn = getValueFromJSON(window, data[action_key]);
                            if (fn) {
                                fn(event, data, _this7);
                            } else {
                                throw new Error("Invalid function: " + data[action_key] + " in View " + _this7.constructor.name);
                            }
                        }
                    });
                    actionElement.dataset['_adharaevent_'] = true;
                };

                var _iteratorNormalCompletion9 = true;
                var _didIteratorError9 = false;
                var _iteratorError9 = undefined;

                try {
                    for (var _iterator9 = onActionElements[Symbol.iterator](), _step9; !(_iteratorNormalCompletion9 = (_step9 = _iterator9.next()).done); _iteratorNormalCompletion9 = true) {
                        var actionElement = _step9.value;

                        var _ret3 = _loop3(actionElement);

                        if (_ret3 === "continue") continue;
                    }
                } catch (err) {
                    _didIteratorError9 = true;
                    _iteratorError9 = err;
                } finally {
                    try {
                        if (!_iteratorNormalCompletion9 && _iterator9.return) {
                            _iterator9.return();
                        }
                    } finally {
                        if (_didIteratorError9) {
                            throw _iteratorError9;
                        }
                    }
                }
            };

            var _arr = ["click", "change"];

            for (var _i = 0; _i < _arr.length; _i++) {
                var action = _arr[_i];
                _loop2(action);
            }
        }
    }, {
        key: "format",
        value: function format(container) {
            //Control the DOM elements after rendering
        }
    }, {
        key: "refresh",
        value: function refresh() {
            Adhara.createView(this);
        }
    }, {
        key: "onDestroy",
        value: function onDestroy() {
            // This method will be called just before destroying the view
        }
    }, {
        key: "destroy",
        value: function destroy() {
            // This method will destroy the view
            this.onDestroy();
            this.trigger("ViewDestroyed");
            this.is_active = false;
            try {
                this.getParentContainerElement().innerHTML = "";
            } catch (e) {/*Do nothing*/}
        }
    }, {
        key: "events",
        get: function get() {
            return [];
        }
    }, {
        key: "template",
        get: function get() {
            if (Adhara.router.getCurrentPageName()) {
                return Adhara.router.getCurrentPageName().replace(/_/g, "-");
            } else {
                var className = this.constructor.name.toLowerCase();
                if (className.endsWith("view")) {
                    return className.slice(0, className.length - 4);
                } else {
                    return className;
                }
            }
        }
    }, {
        key: "errorTemplate",
        get: function get() {
            return this.getCustomTemplate("error") || null;
        }

        /**
         * @method
         * @getter
         * @returns {HandlebarTemplate} structure to be rendered when data is being fetched
         * */

    }, {
        key: "fetchingDataTemplate",
        get: function get() {
            return this.getCustomTemplate("fetching_data") || "";
        }

        /**
         * @method
         * @getter
         * @returns {HandlebarTemplate} structure to be rendered when there is no data available
         * */

    }, {
        key: "noDataTemplate",
        get: function get() {
            return this.getCustomTemplate("no_data") || "";
        }
    }, {
        key: "payload",
        get: function get() {
            return null;
        }
    }, {
        key: "entityConfig",
        get: function get() {
            var entity_name = Adhara.getViewContext(this);
            if (!entity_name) {
                return null;
            }
            return this.formatEntityConfig(Adhara.app.getEntityConfig(entity_name));
        }
    }, {
        key: "state",
        get: function get() {
            return this._state;
        }
    }, {
        key: "errors",
        get: function get() {
            return this._error;
        }

        /**
         * @method
         * @getter
         * @returns {*} View data, that can be consumed by the template.
         * */

    }, {
        key: "data",
        get: function get() {
            return this._data;
        }
    }, {
        key: "parentView",
        get: function get() {
            return this._parentViewInstance || Adhara.getView(Adhara.app.containerView);
        }
    }, {
        key: "contentContainer",
        get: function get() {
            throw new Error("implement this method");
        }

        /**
         * @typedef {Object} SubView - Sub view configuration
         * @property {String} container_selector - container's CSS selector
         * @property {class} view - view class reference of the sub view
         * */

        /**
         * @function
         * @instance
         * @returns Array<SubView>
         * */

    }, {
        key: "subViews",
        get: function get() {
            return [];
        }
    }]);

    return AdharaView;
}(AdharaController);

/**------------------------------------------------------------------------------------------------------------------**/


var AdharaListView = function (_AdharaView) {
    _inherits(AdharaListView, _AdharaView);

    function AdharaListView(parentViewInstance) {
        _classCallCheck(this, AdharaListView);

        var _this8 = _possibleConstructorReturn(this, (AdharaListView.__proto__ || Object.getPrototypeOf(AdharaListView)).call(this, parentViewInstance));

        _this8._page_number = 1;
        return _this8;
    }

    _createClass(AdharaListView, [{
        key: "onPreviousPage",


        /**
         * @function
         * @instance
         * @description listens to the previous page request, and triggers the Page Change Listener
         * */
        value: function onPreviousPage() {
            this._page_number = this.pageNumber - 1;
            this.pageChange();
        }

        /**
         * @function
         * @instance
         * @description listens to the next page request, and triggers the Page Change Listener
         * */

    }, {
        key: "onNextPage",
        value: function onNextPage() {
            this._page_number = this.pageNumber + 1;
            this.pageChange();
        }

        /**
         * @function
         * @instance
         * @description will be called on page change
         * */

    }, {
        key: "pageChange",
        value: function pageChange() {
            this.fetchData();
        }
    }, {
        key: "getPayload",
        value: function getPayload() {
            if (this.isPaginationRequired) {
                return Object.assign({}, _get(AdharaListView.prototype.__proto__ || Object.getPrototypeOf(AdharaListView.prototype), "getPayload", this).call(this), AdharaListView.getPagePayload(this.pageNumber));
            }
            return _get(AdharaListView.prototype.__proto__ || Object.getPrototypeOf(AdharaListView.prototype), "getPayload", this).call(this);
        }
    }, {
        key: "template",
        get: function get() {
            return "adhara-list";
        }

        /**
         * @function
         * @getter
         * @returns {HandlebarTemplate} template of the list structure
         * */

    }, {
        key: "listTemplate",
        get: function get() {
            throw new Error("override `get listTemplate`");
        }

        /**
         * @function
         * @getter
         * @returns {HandlebarTemplate} template of the list header
         * */

    }, {
        key: "headerTemplate",
        get: function get() {
            throw new Error("override `get headerTemplate`");
        }

        /**
         * @method
         * @getter
         * @returns {HandlebarTemplate} template of a list item
         * */

    }, {
        key: "itemTemplate",
        get: function get() {
            throw new Error("override `get itemTemplate`");
        }

        /**
         * @method
         * @getter
         * @returns {String} Table title that is to be rendered
         * */

    }, {
        key: "title",
        get: function get() {
            return "";
        }

        /**
         * @method
         * @getter
         * @returns {Boolean} whether to enable pagination or not
         * */

    }, {
        key: "isPaginationRequired",
        get: function get() {
            return false;
        }

        /**
         * @typedef {Object} ColumnConfig
         * @property {String} key - column key
         * @property {String} name - column name that is to be rendered
         * @property {Boolean} trust_as_html - whether to render as text content or HTML content
         * @example
         * {
         *  key: "column_key",
         *  name: "Column Name",
         *  trust_as_html: false // whether to render HTML as buttons or not
         * }
         * */

        /**
         * @method
         * @getter
         * @returns {Array<ColumnConfig>} list of columns in the format of column configuration
         * */

    }, {
        key: "columns",
        get: function get() {
            return [];
        }
    }, {
        key: "noDataAvailable",
        get: function get() {
            return "No Data Available";
        }

        /**
         * @function
         * @getter
         * @returns {HandlebarTemplate}
         * @description this template contents will be rendered just above the table.
         * */

    }, {
        key: "menuTemplate",
        get: function get() {
            return "adhara-list-menu";
        }

        /**
         * @function
         * @getter
         * @returns {Array<Object>} Buttons to be added in the header
         * @example
         * return [
         *  {
         *      title: "Add new element",
         *      attributes: {
         *          "class": "btn btn-primary",
         *          "data-onclick": "openForm",
         *      }
         *  }
         * ]
         * */

    }, {
        key: "buttons",
        get: function get() {
            return [];
        }

        /**
         * @function
         * @getter
         * @returns {Number} total row count in a page
         * */

    }, {
        key: "rowCount",
        get: function get() {
            return AdharaListView.rowCount;
        }

        /**
         * @function
         * @getter
         * @returns {Number} the start index of the current page
         * */

    }, {
        key: "pageNumber",
        get: function get() {
            return this._page_number;
        }

        /**
         * @function
         * @getter
         * @returns {Boolean} if the current page is the first page
         * */

    }, {
        key: "isFirstPage",
        get: function get() {
            return this._page_number === 1;
        }

        /**
         * @function
         * @getter
         * @returns {Boolean} if the current page is the last page
         * */

    }, {
        key: "isLastPage",
        get: function get() {
            //Assumption: this.data is the only has the data for the current page
            return this.data.length < this.rowCount;
        }
    }]);

    return AdharaListView;
}(AdharaView);

/**
 * @member {Number} default row count of all table instances. Default's 5.
 * Override to increase globally.
 * */


AdharaListView.rowCount = 5;

/**
 * @member {Function} Global function that implements pagePayload pattern for payload on pagination.
 * Recommended to override.
 * */
AdharaListView.getPagePayload = function (page) {
    return { page: page };
};
/**------------------------------------------------------------------------------------------------------------------**/

var AdharaFormView = function (_AdharaView2) {
    _inherits(AdharaFormView, _AdharaView2);

    function AdharaFormView() {
        _classCallCheck(this, AdharaFormView);

        return _possibleConstructorReturn(this, (AdharaFormView.__proto__ || Object.getPrototypeOf(AdharaFormView)).apply(this, arguments));
    }

    _createClass(AdharaFormView, [{
        key: "onSuccess",


        /**
         * @function
         * @instance
         * @param {Object} data - Response Object
         * @description This method will be called on forms's POST/PUT call success
         * */
        value: function onSuccess(data) {
            Toast.success("Success");
        }

        /**
         * @function
         * @instance
         * @param {Object} error - Error Response Object
         * @description This method will be called on forms's POST/PUT failure
         * */

    }, {
        key: "onError",
        value: function onError(error) {
            Toast.error(error);
        }

        /**
         * @function
         * @instance
         * @param {Error} error - Error thrown from validator
         * */

    }, {
        key: "onValidationError",
        value: function onValidationError(error) {
            Toast.error(error.message);
        }

        /**
         * @function
         * @instance
         * @param {Object} data - Form data
         * @returns {Boolean} Whether the data is valid or not
         * */

    }, {
        key: "validate",
        value: function validate(data) {
            var validate_fn = this.formElement.getAttribute('validate-data');
            if (validate_fn) {
                return !!call_fn(validate_fn, data);
            }
        }

        /**
         * @function
         * @instance
         * @param {Object} data - Form data
         * @returns {Object} Modified/Formatted data
         * */

    }, {
        key: "formatData",
        value: function formatData(data) {
            var format_fn = this.formElement.getAttribute('format-data');
            if (format_fn) {
                data = call_fn(format_fn, data);
            }
            return data;
        }

        /**
         * @getter
         * @instance
         * @returns {HTMLFormElement} HTML node which represents the Form element
         * */

    }, {
        key: "getFieldValue",


        /**
         * @funciton
         * @instance
         * @returns {*} Field data
         * */
        value: function getFieldValue(field_name) {
            var field = this.formElement[field_name];
            if (field.type === "number") {
                return +field.value;
            } else if (field.type === "checkbox") {
                return field.checked;
            }
            return field.value;
        }

        /**
         * @funciton
         * @instance
         * @returns {*} Field data
         * */

    }, {
        key: "setFieldValue",
        value: function setFieldValue(field_name, value) {
            var field = this.formElement[field_name];
            if (field.type === "checkbox") {
                field.checked = true;
            } else {
                field.value = value;
            }
        }

        /**
         * @getter
         * @instance
         * @description whether to handle file uploads or not!
         * */

    }, {
        key: "updateFormState",
        value: function updateFormState(submitting) {
            this.formElement.submitting = submitting;
            this.handleSubmitButton();
        }
    }, {
        key: "handleSubmitButton",
        value: function handleSubmitButton() {
            var submit_button = this.submitButton;
            if (!submit_button) {
                console.warn("Unable to discover 'Submit button'. Duplicate submission are not handled.");
                return;
            }
            if (this.formElement.submitting) {
                submit_button.dataset.tosubmit = submit_button.innerHTML;
                submit_button.innerHTML = submit_button.dataset.inprogress || submit_button.dataset.tosubmit;
                submit_button.disabled = true;
            } else {
                submit_button.dataset.inprogress = submit_button.innerHTML;
                submit_button.innerHTML = submit_button.dataset.tosubmit;
                submit_button.disabled = false;
            }
        }

        /**
         * @getter
         * @private
         * @description handle making API call on behalf of the form
         * */

    }, {
        key: "_handleForm",
        value: function _handleForm() {
            var form = this.formElement;
            if (form.submitting) {
                return;
            }
            this.updateFormState(true);
            var apiData = void 0;
            var fileElements = Array.prototype.slice.apply(form.querySelectorAll('input[type="file"]')).filter(function (fileElement) {
                return !fileElement.disabled;
            });
            if (this.handleFileUploads && !!fileElements.length) {
                // ~ if(hasFiles){
                apiData = new FormData(form);
            } else {
                var formData = jQuery(form).serializeArray();
                apiData = {};
                jQuery.each(formData, function (i, fieldData) {
                    apiData[fieldData.name] = fieldData.value;
                });
            }
            try {
                this.validate(apiData);
            } catch (e) {
                return this.onValidationError(e);
            }
            apiData = this.formatData(apiData);
            return this.control(this.method, this.formEntityConfig, apiData);
        }
    }, {
        key: "_format",
        value: function _format(container) {
            var _this10 = this;

            _get(AdharaFormView.prototype.__proto__ || Object.getPrototypeOf(AdharaFormView.prototype), "_format", this).call(this, container);
            if (this.errors) {
                return;
            }
            this.formElement.addEventListener("submit", function (event) {
                event.preventDefault();
                _this10.submit();
            });
        }
    }, {
        key: "submit",
        value: function submit() {
            this._handleForm(this.formElement);
        }
    }, {
        key: "action",


        /**
         * @getter
         * @returns {String|null} action, URL to be called to post data to.
         * */
        get: function get() {
            return this.formElement ? this.formElement.action.split(window.location.host)[1] : "";
        }
    }, {
        key: "method",
        get: function get() {
            return this.formElement ? this.formElement.getAttribute('api-method') || "post" : "post";
        }
    }, {
        key: "formEntityConfig",
        get: function get() {
            var _this11 = this;

            var form = this.formElement;
            return Adhara.app.getEntityConfigFromContext({
                data_config: {
                    url: this.formatURL(this.action, this.getURLPathParams()),
                    allowed_query_types: [this.method],
                    default_query_type: this.method
                },
                processor: {
                    success: function success(query_type, entity_config, response, xhr, pass_over) {
                        if (_this11.clearFormOnSuccess) {
                            form.reset();
                        }
                        _this11.updateFormState(false);
                        _this11.onSuccess(response);
                    },
                    error: function error(query_type, entity_config, _error, xhr, pass_over) {
                        _this11.updateFormState(false);
                        _this11.onError(_error);
                    }
                }
            });
        }

        /**
         * @getter
         * @instance
         * @returns {Boolean} whether to clear form on successful submission or not
         * */

    }, {
        key: "clearFormOnSuccess",
        get: function get() {
            return this.formElement.getAttribute('form-clear') === "true";
        }
    }, {
        key: "formElement",
        get: function get() {
            if (!this._formElement || !document.body.contains(this._formElement)) {
                this._formElement = this.getParentContainerElement().querySelector("form");
            }
            if (!this._formElement) {
                throw new Error("No from element discovered!");
            }
            return this._formElement;
        }
    }, {
        key: "handleFileUploads",
        get: function get() {
            return true;
        }
    }, {
        key: "submitButton",
        get: function get() {
            return this.formElement.querySelector('[type="submit"]');
        }
    }]);

    return AdharaFormView;
}(AdharaView);
/**------------------------------------------------------------------------------------------------------------------**/
/**
 * Created by varun on 19/12/17.
 * TODO : 1. need to take in different table configurations, 2. need a preprocessor for data before drawing table
 */

var AdharaGridView = function (_AdharaListView) {
    _inherits(AdharaGridView, _AdharaListView);

    function AdharaGridView() {
        _classCallCheck(this, AdharaGridView);

        return _possibleConstructorReturn(this, (AdharaGridView.__proto__ || Object.getPrototypeOf(AdharaGridView)).apply(this, arguments));
    }

    _createClass(AdharaGridView, [{
        key: "listTemplate",
        get: function get() {
            return 'adhara-list-grid';
        }
    }, {
        key: "tableClass",
        get: function get() {
            return "table-bordered table-hover";
        }
    }]);

    return AdharaGridView;
}(AdharaListView);
/**------------------------------------------------------------------------------------------------------------------**/


var AdharaTemplateView = function (_AdharaListView2) {
    _inherits(AdharaTemplateView, _AdharaListView2);

    function AdharaTemplateView() {
        _classCallCheck(this, AdharaTemplateView);

        return _possibleConstructorReturn(this, (AdharaTemplateView.__proto__ || Object.getPrototypeOf(AdharaTemplateView)).apply(this, arguments));
    }

    _createClass(AdharaTemplateView, [{
        key: "listTemplate",
        get: function get() {
            return 'adhara-list-template';
        }
    }, {
        key: "headerTemplate",
        get: function get() {
            return 'adhara-list-template-header';
        }
    }, {
        key: "itemTemplate",
        get: function get() {
            return 'adhara-list-template-item';
        }
    }, {
        key: "tableClass",
        get: function get() {
            return "table-bordered table-hover";
        }
    }, {
        key: "containerAttributes",
        get: function get() {
            return {};
        }
    }]);

    return AdharaTemplateView;
}(AdharaListView);
/**------------------------------------------------------------------------------------------------------------------**/


var AdharaCardView = function (_AdharaListView3) {
    _inherits(AdharaCardView, _AdharaListView3);

    function AdharaCardView() {
        _classCallCheck(this, AdharaCardView);

        return _possibleConstructorReturn(this, (AdharaCardView.__proto__ || Object.getPrototypeOf(AdharaCardView)).apply(this, arguments));
    }

    _createClass(AdharaCardView, [{
        key: "template",
        get: function get() {
            return "adhara-card";
        }
    }, {
        key: "containerClass",
        get: function get() {
            return "";
        }
    }, {
        key: "cardSizeClass",
        get: function get() {
            return "col-md-4";
        }
    }, {
        key: "itemTemplate",
        get: function get() {
            return "adhara-card-content";
        }
    }]);

    return AdharaCardView;
}(AdharaListView);
/**------------------------------------------------------------------------------------------------------------------**/
/**
 * Created by varun on 10/2/18.
 */


var AdharaDialogView = function (_AdharaView3) {
    _inherits(AdharaDialogView, _AdharaView3);

    function AdharaDialogView() {
        _classCallCheck(this, AdharaDialogView);

        return _possibleConstructorReturn(this, (AdharaDialogView.__proto__ || Object.getPrototypeOf(AdharaDialogView)).apply(this, arguments));
    }

    _createClass(AdharaDialogView, [{
        key: "onInit",
        value: function onInit() {
            this._modalId = "adhara-dialog-id-" + Date.now();
            this.destroy = this.destroy.bind(this);
        }
    }, {
        key: "getParentContainerElement",
        value: function getParentContainerElement() {
            // return document.querySelector('.adhara-dialog')
            return document.querySelector('#' + this.wrapperId);
        }
    }, {
        key: "render",
        value: function render() {
            var dialog_elements = document.querySelectorAll('#' + this.modalId);
            if (dialog_elements && dialog_elements.length) {
                for (var i = 0; i < dialog_elements.length; i++) {
                    dialog_elements[i].remove();
                }
            }
            var wrapper = document.createElement('div');
            wrapper.id = this.wrapperId;
            wrapper.classList.add('adhara-dialog');
            document.querySelector('body').appendChild(wrapper);
            _get(AdharaDialogView.prototype.__proto__ || Object.getPrototypeOf(AdharaDialogView.prototype), "render", this).call(this);
        }
    }, {
        key: "format",
        value: function format() {
            if (this.isAutoShow) {
                this.show();
            }
        }
    }, {
        key: "show",
        value: function show() {
            var _this16 = this;

            this.modalElement.modal('show');
            if (this.destroyOnClose) {
                setTimeout(function () {
                    _this16.modalElement.on('hidden.bs.modal', _this16.destroy);
                }, 0);
            }
        }
    }, {
        key: "hide",
        value: function hide() {
            //Note: destroy's if destroyOnClose is enabled. See event listener binding in show() method.
            this.modalElement.modal('hide');
        }
    }, {
        key: "onDestroy",
        value: function onDestroy() {
            var wrapper_element = document.getElementById(this.wrapperId);
            if (wrapper_element) {
                wrapper_element.remove();
            }
        }
    }, {
        key: "template",
        get: function get() {
            return "adhara-dialog";
        }
    }, {
        key: "titleTemplate",
        get: function get() {
            return 'Adhara Dialog';
        }
    }, {
        key: "messageTemplate",
        get: function get() {
            return 'Adhara Dialog Message!';
        }
    }, {
        key: "buttons",
        get: function get() {
            return [{
                attributes: {
                    "type": "button",
                    "class": "btn btn-secondary",
                    "data-dismiss": "modal"
                },
                text: 'Close'
            }];
        }
    }, {
        key: "modalId",
        get: function get() {
            return this._modalId;
        }
    }, {
        key: "wrapperId",
        get: function get() {
            return this.modalId + "-wrapper";
        }
    }, {
        key: "modalElement",
        get: function get() {
            if (!this._modalElemnt || !document.body.contains(this._modalElemnt[0])) {
                this._modalElemnt = $('#' + this.modalId); //Using jQ as that is mostly used in this view
            }
            return this._modalElemnt;
        }
    }, {
        key: "isAutoShow",
        get: function get() {
            return true;
        }
    }, {
        key: "destroyOnClose",
        get: function get() {
            return true;
        }
    }]);

    return AdharaDialogView;
}(AdharaView);
/**------------------------------------------------------------------------------------------------------------------**/


var AdharaTabView = function (_AdharaView4) {
    _inherits(AdharaTabView, _AdharaView4);

    function AdharaTabView() {
        _classCallCheck(this, AdharaTabView);

        return _possibleConstructorReturn(this, (AdharaTabView.__proto__ || Object.getPrototypeOf(AdharaTabView)).apply(this, arguments));
    }

    _createClass(AdharaTabView, [{
        key: "onTabShow",
        value: function onTabShow() {}
    }, {
        key: "onTabClick",
        value: function onTabClick() {}
    }, {
        key: "onTabChange",
        value: function onTabChange() {}

        /*onShow(){
          }*/

    }, {
        key: "format",
        value: function format() {}
    }, {
        key: "renderSubViews",
        value: function renderSubViews() {
            Adhara.createView(Adhara.getView(this.currentTab.view, this));
        }
    }, {
        key: "template",
        get: function get() {
            return "adhara-tab-view";
        }
    }, {
        key: "containerClass",
        get: function get() {
            return 'nav nav-pills';
        }
    }, {
        key: "tabNavClass",
        get: function get() {
            return 'custom-tabs-line tabs-line-bottom left-aligned';
        }
    }, {
        key: "tabListClass",
        get: function get() {
            return "nav";
        }
    }, {
        key: "tabs",
        get: function get() {
            return [];
        }

        /**
         * @getter
         * @instance
         * @returns {Object} Active tab configuration
         * */

    }, {
        key: "currentTab",
        get: function get() {
            var current_tab_link_from_url = Adhara.router.getCurrentURL();
            var current_Tab = this.tabs.filter(function (tab) {
                return tab.link === current_tab_link_from_url;
            });
            return current_Tab.length ? current_Tab[0] : this.tabs[0];
        }
    }, {
        key: "tabsList",
        get: function get() {
            var current_tab_id = this.currentTab.id;
            return this.tabs.map(function (tab) {
                tab.className = tab.id === current_tab_id ? "active" : "";
                return tab;
            });
        }
    }, {
        key: "nextSelector",
        get: function get() {
            return '.btn-next';
        }
    }, {
        key: "previousSelector",
        get: function get() {
            return '.btn-previous';
        }
    }, {
        key: "contentContainer",
        get: function get() {
            return ".tab-content .tab-pane";
        }
    }]);

    return AdharaTabView;
}(AdharaView);
/**------------------------------------------------------------------------------------------------------------------**/


var AdharaContainerView = function (_AdharaView5) {
    _inherits(AdharaContainerView, _AdharaView5);

    function AdharaContainerView() {
        _classCallCheck(this, AdharaContainerView);

        return _possibleConstructorReturn(this, (AdharaContainerView.__proto__ || Object.getPrototypeOf(AdharaContainerView)).apply(this, arguments));
    }

    _createClass(AdharaContainerView, [{
        key: "_getParentContainer",
        value: function _getParentContainer() {
            return Adhara.app.DOMSelector;
        }
    }, {
        key: "template",
        get: function get() {
            return "container";
        }
    }, {
        key: "contentContainer",
        get: function get() {
            return "main";
        }
    }]);

    return AdharaContainerView;
}(AdharaView);
/**------------------------------------------------------------------------------------------------------------------**/


var AdharaDialogFormView = function (_MutateViews) {
    _inherits(AdharaDialogFormView, _MutateViews);

    function AdharaDialogFormView() {
        _classCallCheck(this, AdharaDialogFormView);

        return _possibleConstructorReturn(this, (AdharaDialogFormView.__proto__ || Object.getPrototypeOf(AdharaDialogFormView)).apply(this, arguments));
    }

    _createClass(AdharaDialogFormView, [{
        key: "primaryAction",
        value: function primaryAction() {
            this.submit();
        }
    }, {
        key: "primaryActionTitle",
        get: function get() {
            return 'Save';
        }
    }, {
        key: "submitButton",
        get: function get() {
            return this.getParentContainerElement().querySelector('button.btn-primary');
        }
    }, {
        key: "buttons",
        get: function get() {
            if (this.primaryActionTitle) {
                return [_get(AdharaDialogFormView.prototype.__proto__ || Object.getPrototypeOf(AdharaDialogFormView.prototype), "buttons", this)[0], {
                    attributes: {
                        "type": "button",
                        "class": "btn btn-primary",
                        "data-onclick": "primaryAction"
                    },
                    text: this.primaryActionTitle
                }];
            } else {
                return _get(AdharaDialogFormView.prototype.__proto__ || Object.getPrototypeOf(AdharaDialogFormView.prototype), "buttons", this);
            }
        }
    }]);

    return AdharaDialogFormView;
}(MutateViews(AdharaFormView, AdharaDialogView));
/**------------------------------------------------------------------------------------------------------------------**/
/**
 * @class
 * @abstract
 * @classdesc operations that can be performed on a specific dataset whose name must be passed as initializer argument
 * */


var AdharaStorage = function () {
    function AdharaStorage() {
        _classCallCheck(this, AdharaStorage);
    }

    _createClass(AdharaStorage, [{
        key: "select",


        /**
         * @function
         * @instance
         * @param {String} database_name - name of the database
         * @param {String} dataset_name - name/key_name of the dataset on which operations are to be carried out
         * */
        value: function select(database_name, dataset_name) {
            throw new Error("return a StorageOperator instance from this method. create a `class xStorageOperator extends AdharaStorageOperator{...}`");
        }
    }]);

    return AdharaStorage;
}();

/**
 * @class
 * @abstract
 * @classdesc operations that can be performed on a specific dataset whose name must be passed as initializer argument
 * */


var AdharaStorageOperator = function () {

    /**
     * @constructor
     * @param {String} dataset - name/key_name of the dataset on which operations are to be carried out
     * */
    function AdharaStorageOperator(dataset) {
        _classCallCheck(this, AdharaStorageOperator);

        this.dataset = dataset;
    }

    /**
     * @function
     * @instance
     * @abstract
     * @returns {Promise}
     * */


    _createClass(AdharaStorageOperator, [{
        key: "store",
        value: function store(key, value) {
            throw new Error("Override `store` method");
        }

        /**
         * @function
         * @instance
         * @abstract
         * @returns {Promise}
         * */

    }, {
        key: "retrieve",
        value: function retrieve(key) {
            throw new Error("Override `retrieve` method");
        }

        /**
         * @function
         * @instance
         * @abstract
         * @returns {Promise}
         * */

    }, {
        key: "remove",
        value: function remove(key) {
            throw new Error("Override `remove` method");
        }

        /**
         * @function
         * @instance
         * @returns {Promise}
         * */

    }, {
        key: "removeAll",
        value: function removeAll() {
            return this.removeMultiple(function () {
                return true;
            });
        }

        /**
         * @function
         * @instance
         * @param {Function} filter - A function whose return value is interpreted as a boolean. return true, will remove the entry, else do nothing
         * @returns {Promise}
         * */

    }, {
        key: "removeMultiple",
        value: function removeMultiple(filter) {
            var _this20 = this;

            return this.keys().then(function () {
                var _ref4 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee(keys) {
                    var key_promises, _iteratorNormalCompletion10, _didIteratorError10, _iteratorError10, _iterator10, _step10, key, data, data_promises, i, datum, _key4;

                    return regeneratorRuntime.wrap(function _callee$(_context) {
                        while (1) {
                            switch (_context.prev = _context.next) {
                                case 0:
                                    key_promises = [];
                                    _iteratorNormalCompletion10 = true;
                                    _didIteratorError10 = false;
                                    _iteratorError10 = undefined;
                                    _context.prev = 4;
                                    _iterator10 = keys[Symbol.iterator]();

                                case 6:
                                    if (_iteratorNormalCompletion10 = (_step10 = _iterator10.next()).done) {
                                        _context.next = 16;
                                        break;
                                    }

                                    key = _step10.value;
                                    _context.t0 = key_promises;
                                    _context.next = 11;
                                    return _this20.retrieve(key);

                                case 11:
                                    _context.t1 = _context.sent;

                                    _context.t0.push.call(_context.t0, _context.t1);

                                case 13:
                                    _iteratorNormalCompletion10 = true;
                                    _context.next = 6;
                                    break;

                                case 16:
                                    _context.next = 22;
                                    break;

                                case 18:
                                    _context.prev = 18;
                                    _context.t2 = _context["catch"](4);
                                    _didIteratorError10 = true;
                                    _iteratorError10 = _context.t2;

                                case 22:
                                    _context.prev = 22;
                                    _context.prev = 23;

                                    if (!_iteratorNormalCompletion10 && _iterator10.return) {
                                        _iterator10.return();
                                    }

                                case 25:
                                    _context.prev = 25;

                                    if (!_didIteratorError10) {
                                        _context.next = 28;
                                        break;
                                    }

                                    throw _iteratorError10;

                                case 28:
                                    return _context.finish(25);

                                case 29:
                                    return _context.finish(22);

                                case 30:
                                    _context.next = 32;
                                    return Promise.all(key_promises);

                                case 32:
                                    data = _context.sent;
                                    data_promises = [];

                                    for (i = 0; i < data.length; i++) {
                                        datum = data[i];
                                        _key4 = keys[i];

                                        if (!!filter(_key4, datum)) {
                                            data_promises.push(_this20.remove(_key4));
                                        }
                                    }
                                    _context.next = 37;
                                    return Promise.all(data_promises);

                                case 37:
                                    return _context.abrupt("return", _context.sent);

                                case 38:
                                case "end":
                                    return _context.stop();
                            }
                        }
                    }, _callee, _this20, [[4, 18, 22, 30], [23,, 25, 29]]);
                }));

                return function (_x) {
                    return _ref4.apply(this, arguments);
                };
            }());
        }

        /**
         * @function
         * @instance
         * @abstract
         * @returns {Promise}
         * */

    }, {
        key: "keys",
        value: function keys() {
            throw new Error("Override `keys` method");
        }
    }]);

    return AdharaStorageOperator;
}();
/**------------------------------------------------------------------------------------------------------------------**/


var StorageSelector = function () {
    function StorageSelector() {
        _classCallCheck(this, StorageSelector);
    }

    _createClass(StorageSelector, null, [{
        key: "select",
        value: function select() {
            if (window.hasOwnProperty("indexedDB")) {
                return AdharaDBStorage;
            } else {
                return AdharaCacheStorage;
            }
        }
    }]);

    return StorageSelector;
}();
/**------------------------------------------------------------------------------------------------------------------**/


var AdharaDBStorage = function (_AdharaStorage) {
    _inherits(AdharaDBStorage, _AdharaStorage);

    function AdharaDBStorage() {
        _classCallCheck(this, AdharaDBStorage);

        // promise based Async wrapper for indexedDB

        var _this21 = _possibleConstructorReturn(this, (AdharaDBStorage.__proto__ || Object.getPrototypeOf(AdharaDBStorage)).call(this));

        function promisifyRequest(request) {
            return new Promise(function (resolve, reject) {
                request.onsuccess = function () {
                    resolve(request.result);
                };
                request.onerror = function () {
                    reject(request.error);
                };
            });
        }

        function promisifyRequestCall(obj, method, args) {
            var request = void 0;
            var p = new Promise(function (resolve, reject) {
                request = obj[method].apply(obj, args);
                promisifyRequest(request).then(resolve, reject);
            });
            p.request = request;
            return p;
        }

        function promisifyCursorRequestCall(obj, method, args) {
            var p = promisifyRequestCall(obj, method, args);
            return p.then(function (value) {
                if (!value) {
                    return;
                }
                return new Cursor(value, p.request);
            });
        }

        function proxyProperties(ProxyClass, targetProp, properties) {
            properties.forEach(function (prop) {
                Object.defineProperty(ProxyClass.prototype, prop, {
                    get: function get() {
                        return this[targetProp][prop];
                    },
                    set: function set(val) {
                        this[targetProp][prop] = val;
                    }
                });
            });
        }

        function proxyRequestMethods(ProxyClass, targetProp, Constructor, properties) {
            properties.forEach(function (prop) {
                if (!(prop in Constructor.prototype)) {
                    return;
                }
                ProxyClass.prototype[prop] = function () {
                    return promisifyRequestCall(this[targetProp], prop, arguments);
                };
            });
        }

        function proxyMethods(ProxyClass, targetProp, Constructor, properties) {
            properties.forEach(function (prop) {
                if (!(prop in Constructor.prototype)) {
                    return;
                }
                ProxyClass.prototype[prop] = function () {
                    return this[targetProp][prop].apply(this[targetProp], arguments);
                };
            });
        }

        function proxyCursorRequestMethods(ProxyClass, targetProp, Constructor, properties) {
            properties.forEach(function (prop) {
                if (!(prop in Constructor.prototype)) {
                    return;
                }
                ProxyClass.prototype[prop] = function () {
                    return promisifyCursorRequestCall(this[targetProp], prop, arguments);
                };
            });
        }

        function Index(index) {
            this._index = index;
        }

        proxyProperties(Index, '_index', ['name', 'keyPath', 'multiEntry', 'unique']);

        proxyRequestMethods(Index, '_index', IDBIndex, ['get', 'getKey', 'getAll', 'getAllKeys', 'count']);

        proxyCursorRequestMethods(Index, '_index', IDBIndex, ['openCursor', 'openKeyCursor']);

        function Cursor(cursor, request) {
            this._cursor = cursor;
            this._request = request;
        }

        proxyProperties(Cursor, '_cursor', ['direction', 'key', 'primaryKey', 'value']);

        proxyRequestMethods(Cursor, '_cursor', IDBCursor, ['update', 'delete']);

        // proxy 'next' methods
        ['advance', 'continue', 'continuePrimaryKey'].forEach(function (methodName) {
            if (!(methodName in IDBCursor.prototype)) {
                return;
            }
            Cursor.prototype[methodName] = function () {
                var cursor = this;
                var args = arguments;
                return Promise.resolve().then(function () {
                    cursor._cursor[methodName].apply(cursor._cursor, args);
                    return promisifyRequest(cursor._request).then(function (value) {
                        return value ? new Cursor(value, cursor._request) : null;
                    });
                });
            };
        });

        function ObjectStore(store) {
            this._store = store;
        }

        ObjectStore.prototype.createIndex = function () {
            return new Index(this._store.createIndex.apply(this._store, arguments));
        };

        ObjectStore.prototype.index = function () {
            return new Index(this._store.index.apply(this._store, arguments));
        };

        proxyProperties(ObjectStore, '_store', ['name', 'keyPath', 'indexNames', 'autoIncrement']);

        proxyRequestMethods(ObjectStore, '_store', IDBObjectStore, ['put', 'add', 'delete', 'clear', 'get', 'getAll', 'getKey', 'getAllKeys', 'count']);

        proxyCursorRequestMethods(ObjectStore, '_store', IDBObjectStore, ['openCursor', 'openKeyCursor']);

        proxyMethods(ObjectStore, '_store', IDBObjectStore, ['deleteIndex']);

        function Transaction(idbTransaction) {
            this._tx = idbTransaction;
            this.complete = new Promise(function (resolve, reject) {
                idbTransaction.oncomplete = function () {
                    resolve();
                };
                idbTransaction.onerror = function () {
                    reject(idbTransaction.error);
                };
                idbTransaction.onabort = function () {
                    reject(idbTransaction.error);
                };
            });
        }

        Transaction.prototype.objectStore = function () {
            return new ObjectStore(this._tx.objectStore.apply(this._tx, arguments));
        };

        proxyProperties(Transaction, '_tx', ['objectStoreNames', 'mode']);

        proxyMethods(Transaction, '_tx', IDBTransaction, ['abort']);

        function UpgradeDB(db, oldVersion, transaction) {
            this._db = db;
            this.oldVersion = oldVersion;
            this.transaction = new Transaction(transaction);
        }

        UpgradeDB.prototype.createObjectStore = function () {
            return new ObjectStore(this._db.createObjectStore.apply(this._db, arguments));
        };

        proxyProperties(UpgradeDB, '_db', ['name', 'version', 'objectStoreNames']);

        proxyMethods(UpgradeDB, '_db', IDBDatabase, ['deleteObjectStore', 'close']);

        function DB(db) {
            this._db = db;
        }

        DB.prototype.transaction = function () {
            return new Transaction(this._db.transaction.apply(this._db, arguments));
        };

        proxyProperties(DB, '_db', ['name', 'version', 'objectStoreNames']);

        proxyMethods(DB, '_db', IDBDatabase, ['close']);

        // Add cursor iterators
        ['openCursor', 'openKeyCursor'].forEach(function (funcName) {
            [ObjectStore, Index].forEach(function (Constructor) {
                Constructor.prototype[funcName.replace('open', 'iterate')] = function () {
                    var args = Array.prototype.slice.call(arguments);
                    var callback = args[args.length - 1];
                    var nativeObject = this._store || this._index;
                    var request = nativeObject[funcName].apply(nativeObject, args.slice(0, -1));
                    request.onsuccess = function () {
                        callback(request.result);
                    };
                };
            });
        });

        // polyfill getAll
        [Index, ObjectStore].forEach(function (Constructor) {
            if (Constructor.prototype.getAll) {
                return;
            }
            Constructor.prototype.getAll = function (query, count) {
                var instance = this;
                var items = [];
                return new Promise(function (resolve) {
                    instance.iterateCursor(query, function (cursor) {
                        if (!cursor) {
                            resolve(items);
                            return;
                        }
                        items.push(cursor.value);

                        if (count !== undefined && items.length === count) {
                            resolve(items);
                            return;
                        }
                        cursor["continue"]();
                    });
                });
            };
        });

        var idb = {
            open: function open(name, version, upgradeCallback) {
                var p = promisifyRequestCall(indexedDB, 'open', [name, version]);
                var request = p.request;

                request.onupgradeneeded = function (event) {
                    if (upgradeCallback) {
                        upgradeCallback(new UpgradeDB(request.result, event.oldVersion, request.transaction));
                    }
                };

                return p.then(function (db) {
                    return new DB(db);
                });
            },
            del: function del(name) {
                return promisifyRequestCall(indexedDB, 'deleteDatabase', [name]);
            }
        };

        _this21.dbs = {};

        var _loop4 = function _loop4(db_config) {
            _this21.dbs[db_config.id] = idb.open(db_config.name, db_config.version, function (upgradeDb) {
                return _this21.createObjectStores(upgradeDb, db_config);
            });
        };

        var _iteratorNormalCompletion11 = true;
        var _didIteratorError11 = false;
        var _iteratorError11 = undefined;

        try {
            for (var _iterator11 = Adhara.app.DBConfig[Symbol.iterator](), _step11; !(_iteratorNormalCompletion11 = (_step11 = _iterator11.next()).done); _iteratorNormalCompletion11 = true) {
                var db_config = _step11.value;

                _loop4(db_config);
            }
        } catch (err) {
            _didIteratorError11 = true;
            _iteratorError11 = err;
        } finally {
            try {
                if (!_iteratorNormalCompletion11 && _iterator11.return) {
                    _iterator11.return();
                }
            } finally {
                if (_didIteratorError11) {
                    throw _iteratorError11;
                }
            }
        }

        return _this21;
    }

    _createClass(AdharaDBStorage, [{
        key: "createObjectStores",
        value: function createObjectStores(upgradeDb, db_config) {
            try {
                var db_schema = db_config.schema;
                for (var table_name in db_schema) {
                    if (db_schema.hasOwnProperty(table_name)) {
                        if (!upgradeDb.objectStoreNames.contains(table_name)) {
                            //If table doesn't exist already
                            var os = upgradeDb.createObjectStore(table_name, db_schema[table_name]);
                            //indexes = [
                            // // See for more details https://developers.google.com/web/ilt/pwa/working-with-indexeddb#defining_indexes
                            //      {
                            //          name: "title-index",
                            //          property: "title",
                            //          options: {
                            //              unique: true
                            //          }
                            //      }
                            //      {
                            //          name: "gender-index",
                            //          property: "gender",
                            //          options: {
                            //              unique: false
                            //          }
                            //      }
                            // ]
                            var _iteratorNormalCompletion12 = true;
                            var _didIteratorError12 = false;
                            var _iteratorError12 = undefined;

                            try {
                                for (var _iterator12 = db_schema[table_name].indexes[Symbol.iterator](), _step12; !(_iteratorNormalCompletion12 = (_step12 = _iterator12.next()).done); _iteratorNormalCompletion12 = true) {
                                    var index = _step12.value;

                                    os.createIndex(index.name, index.property, index.options);
                                }
                            } catch (err) {
                                _didIteratorError12 = true;
                                _iteratorError12 = err;
                            } finally {
                                try {
                                    if (!_iteratorNormalCompletion12 && _iterator12.return) {
                                        _iterator12.return();
                                    }
                                } finally {
                                    if (_didIteratorError12) {
                                        throw _iteratorError12;
                                    }
                                }
                            }
                        }
                    }
                }
                return true;
            } catch (e) {
                return false;
            }
        }
    }, {
        key: "select",
        value: function select(database_name, dataset_name) {
            return new DBStorageOperator(dataset_name, this.dbs[database_name]);
        }
    }]);

    return AdharaDBStorage;
}(AdharaStorage);

var DBStorageOperator = function (_AdharaStorageOperato) {
    _inherits(DBStorageOperator, _AdharaStorageOperato);

    function DBStorageOperator(dataset, db) {
        _classCallCheck(this, DBStorageOperator);

        var _this22 = _possibleConstructorReturn(this, (DBStorageOperator.__proto__ || Object.getPrototypeOf(DBStorageOperator)).call(this, dataset));

        _this22.db = db;
        return _this22;
    }

    _createClass(DBStorageOperator, [{
        key: "retrieve",
        value: function retrieve(key) {
            var _this23 = this;

            return this.db.then(function (db) {
                return db.transaction(_this23.dataset).objectStore(_this23.dataset).get(key);
            });
        }
    }, {
        key: "store",
        value: function store(key, val) {
            var _this24 = this;

            if (!val) {
                val = key;
                key = undefined;
            }
            return this.db.then(function (db) {
                var tx = db.transaction(_this24.dataset, 'readwrite');
                tx.objectStore(_this24.dataset).put(val, key);
                return tx.complete;
            });
        }
    }, {
        key: "remove",
        value: function remove(key) {
            var _this25 = this;

            return this.db.then(function (db) {
                var tx = db.transaction(_this25.dataset, 'readwrite');
                tx.objectStore(_this25.dataset).delete(key);
                return tx.complete;
            });
        }
    }, {
        key: "removeAll",
        value: function removeAll() {
            var _this26 = this;

            return this.db.then(function (db) {
                var tx = db.transaction(_this26.dataset, 'readwrite');
                tx.objectStore(_this26.dataset).clear();
                return tx.complete;
            });
        }
    }, {
        key: "keys",
        value: function keys() {
            var _this27 = this;

            return this.db.then(function (db) {
                var tx = db.transaction(_this27.dataset);
                var keys = [];
                var store = tx.objectStore(_this27.dataset);
                // This would be store.getAllKeys(), but it isn't supported by Edge or Safari.
                // openKeyCursor isn't supported by Safari, so we fall back
                (store.iterateKeyCursor || store.iterateCursor).call(store, function (cursor) {
                    if (!cursor) return;
                    keys.push(cursor.key);
                    cursor.continue();
                });

                return tx.complete.then(function () {
                    return keys;
                });
            });
        }
    }]);

    return DBStorageOperator;
}(AdharaStorageOperator);
/**------------------------------------------------------------------------------------------------------------------**/


var AdharaCacheStorage = function (_AdharaStorage2) {
    _inherits(AdharaCacheStorage, _AdharaStorage2);

    function AdharaCacheStorage() {
        _classCallCheck(this, AdharaCacheStorage);

        var _this28 = _possibleConstructorReturn(this, (AdharaCacheStorage.__proto__ || Object.getPrototypeOf(AdharaCacheStorage)).call(this));

        _this28.dbs = AdharaCacheStorageOperator.cache;
        var _iteratorNormalCompletion13 = true;
        var _didIteratorError13 = false;
        var _iteratorError13 = undefined;

        try {
            for (var _iterator13 = Adhara.app.DBConfig[Symbol.iterator](), _step13; !(_iteratorNormalCompletion13 = (_step13 = _iterator13.next()).done); _iteratorNormalCompletion13 = true) {
                var db_config = _step13.value;

                if (!_this28.dbs[db_config.id]) {
                    _this28.dbs[db_config.id] = {};
                }
            }
        } catch (err) {
            _didIteratorError13 = true;
            _iteratorError13 = err;
        } finally {
            try {
                if (!_iteratorNormalCompletion13 && _iterator13.return) {
                    _iterator13.return();
                }
            } finally {
                if (_didIteratorError13) {
                    throw _iteratorError13;
                }
            }
        }

        return _this28;
    }

    _createClass(AdharaCacheStorage, [{
        key: "select",
        value: function select(database_name, dataset_name) {
            return new AdharaCacheStorageOperator(dataset_name, this.dbs[database_name]);
        }
    }]);

    return AdharaCacheStorage;
}(AdharaStorage);

var AdharaCacheStorageOperator = function (_AdharaStorageOperato2) {
    _inherits(AdharaCacheStorageOperator, _AdharaStorageOperato2);

    function AdharaCacheStorageOperator(dataset, db) {
        _classCallCheck(this, AdharaCacheStorageOperator);

        var _this29 = _possibleConstructorReturn(this, (AdharaCacheStorageOperator.__proto__ || Object.getPrototypeOf(AdharaCacheStorageOperator)).call(this, dataset));

        _this29.db = db;
        return _this29;
    }

    _createClass(AdharaCacheStorageOperator, [{
        key: "store",
        value: function store(key, value) {
            return new Promise(function (resolve, reject) {
                AdharaCacheStorageOperator.cache[key] = value;
                resolve();
            });
        }
    }, {
        key: "retrieve",
        value: function retrieve(key) {
            return new Promise(function (resolve, reject) {
                resolve(AdharaCacheStorageOperator.cache[key]);
            });
        }
    }, {
        key: "remove",
        value: function remove(key) {
            return new Promise(function (resolve, reject) {
                delete AdharaCacheStorageOperator.cache[key];
                resolve();
            });
        }
    }, {
        key: "removeAll",
        value: function removeAll() {
            return new Promise(function (resolve, reject) {
                AdharaCacheStorageOperator.cache = {};
                resolve();
            });
        }
    }, {
        key: "keys",
        value: function keys() {
            return new Promise(function (resolve, reject) {
                resolve(Object.keys(AdharaCacheStorageOperator.cache));
            });
        }
    }]);

    return AdharaCacheStorageOperator;
}(AdharaStorageOperator);

AdharaCacheStorageOperator.cache = {};
/**------------------------------------------------------------------------------------------------------------------**/

var DataInterface = function (_StorageSelector$sele) {
    _inherits(DataInterface, _StorageSelector$sele);

    function DataInterface() {
        _classCallCheck(this, DataInterface);

        var _this30 = _possibleConstructorReturn(this, (DataInterface.__proto__ || Object.getPrototypeOf(DataInterface)).call(this));

        _this30.config = Adhara.app._diConfig;
        _this30.request_queue = {}; // for NON-get requests
        _this30.rem_que = {};
        _this30.db_table = _this30.select("default", _this30.config.http_cache_table);
        _this30.initDependencies();
        return _this30;
    }

    // local helpers ...


    _createClass(DataInterface, [{
        key: "makeServiceCall",
        value: function makeServiceCall(url, method_name, data, controller) {
            data = Adhara.app.requestMiddleWare(url, method_name, data);
            return new Promise(function (resolve, reject) {
                // make API call and fetch data OR fetch from client storage, pass on_success and on_failure to the thing
                controller[method_name]({
                    url: url, data: data, crossDomain: true, xhrFields: { "withCredentials": true },
                    beforeSend: function beforeSend(xhr) {
                        xhr.withCredentials = true;
                    },
                    success: function success(response, xhr) {
                        resolve({ response: response, xhr: xhr });
                    },
                    failure: function failure(error, xhr) {
                        reject({ error: error, xhr: xhr });
                    }
                });
            });
        }
    }, {
        key: "handle_bulk_config",
        value: function handle_bulk_config(query_type, entity_configs, data, opts) {
            if (['get', 'get_list'].indexOf(query_type) < 0) return;
            var num_reqs = entity_configs.length,
                result = new Array(num_reqs),
                entry_count = 0;
            var caller_viewable = opts.caller_view && (opts.caller_view instanceof AdharaView || opts.caller_view.hasOwnProperty('handleDataChange') && typeof opts.caller_view.handleDataChange === 'function');
            for (var i = 0; i < num_reqs; i++) {
                var dummy_entity_config = Object.assign({}, entity_configs[i]);
                dummy_entity_config.processor = function IIFE(scoped_i) {
                    return {
                        success: function success(query_type, entity_config, response, xhr) {
                            var processed_data = processor_helper.get_basic_processed_data(query_type, entity_config, response, xhr);
                            result[scoped_i] = processed_data;
                            entry_count++;
                            if (caller_viewable) {
                                if (entry_count === num_reqs) {
                                    opts.caller_view.handleDataChange(result);
                                }
                            } else {
                                Adhara.configUtils.getViewInstance(entity_config).handleDataChange(processed_data);
                            }
                        },
                        error: function error(query_type, entity_config, _error2, xhr) {
                            var processed_data = processor_helper.get_basic_processed_data(query_type, entity_config, _error2, xhr);
                            result[scoped_i] = processed_data;
                            entry_count++;
                            if (caller_viewable) {
                                if (entry_count === num_reqs) {
                                    opts.caller_view.handleDataChange(result);
                                }
                            } else {
                                Adhara.configUtils.getViewInstance(entity_config).handleDataError(processed_data);
                            }
                        }
                    };
                }(i);
                this.enqueue(query_type, dummy_entity_config, data, opts);
            }
        }
    }, {
        key: "handle_batch_enqueue",
        value: function handle_batch_enqueue(entity_config) {
            var batch_result_map = {};
            var data_config = Adhara.configUtils.getDataConfig(entity_config);
            var batch_calls = data_config.batch_data_override;
            function check_fills(obj) {
                for (var k in obj) {
                    if (obj.hasOwnProperty(k)) {
                        if (typeof obj[k] === 'undefined') {
                            return false;
                        }
                    }
                }
                return true;
            }
            function publishToView(final_batch_data) {
                Adhara.configUtils.getViewInstance(entity_config).handleBatchData(final_batch_data);
            }
            var batch_processor = { // custom processor for batching requests
                success: function success(query_type, entity_config, response, xhr) {
                    var blob = Adhara.configUtils.getBlobClass(entity_config);
                    var data_config = Adhara.configUtils.getDataConfig(entity_config);
                    var processed_data = void 0;
                    if (query_type === "get_list") {
                        processed_data = [];
                        var _iteratorNormalCompletion14 = true;
                        var _didIteratorError14 = false;
                        var _iteratorError14 = undefined;

                        try {
                            for (var _iterator14 = response[Symbol.iterator](), _step14; !(_iteratorNormalCompletion14 = (_step14 = _iterator14.next()).done); _iteratorNormalCompletion14 = true) {
                                var datum = _step14.value;

                                processed_data.push(new blob(datum));
                            }
                        } catch (err) {
                            _didIteratorError14 = true;
                            _iteratorError14 = err;
                        } finally {
                            try {
                                if (!_iteratorNormalCompletion14 && _iterator14.return) {
                                    _iterator14.return();
                                }
                            } finally {
                                if (_didIteratorError14) {
                                    throw _iteratorError14;
                                }
                            }
                        }
                    } else {
                        processed_data = new blob(response);
                    }
                    batch_result_map[data_config.identifier] = { success: processed_data };
                    if (check_fills(batch_result_map)) publishToView(batch_result_map);
                    processor_helper.on_success_common(query_type, entity_config, response, xhr);
                },
                error: function error(query_type, entity_config, _error3, xhr) {
                    var data_config = Adhara.configUtils.getDataConfig(entity_config);
                    batch_result_map[data_config.identifier] = { error: _error3 };
                    if (check_fills(batch_result_map)) publishToView(batch_result_map);
                }
            };
            var _iteratorNormalCompletion15 = true;
            var _didIteratorError15 = false;
            var _iteratorError15 = undefined;

            try {
                for (var _iterator15 = batch_calls[Symbol.iterator](), _step15; !(_iteratorNormalCompletion15 = (_step15 = _iterator15.next()).done); _iteratorNormalCompletion15 = true) {
                    var batch_call = _step15.value;

                    batch_result_map[batch_call.identifier] = undefined;
                    //TODO add stuff to the cline db after query abd before querying teh ckiebt DV befire quertubg!
                    //add stuff to the client db after query and before querying the client DB
                    var dummy_entity_config = {
                        data_config: {
                            url: batch_call.url,
                            _url: batch_call.url,
                            identifier: batch_call.identifier,
                            default_query_type: batch_call.query_type || 'get',
                            allowed_query_types: ["get", "get_list"],
                            reuse: batch_call.reuse,
                            blob: batch_call.blob
                        },
                        view: Adhara.configUtils.getViewInstance(entity_config),
                        processor: batch_processor,
                        controller: Adhara.configUtils.getController(entity_config)
                    };
                    this.enqueue(undefined, dummy_entity_config, undefined);
                }
            } catch (err) {
                _didIteratorError15 = true;
                _iteratorError15 = err;
            } finally {
                try {
                    if (!_iteratorNormalCompletion15 && _iterator15.return) {
                        _iterator15.return();
                    }
                } finally {
                    if (_didIteratorError15) {
                        throw _iteratorError15;
                    }
                }
            }
        }
    }, {
        key: "isValidStorageData",
        value: function isValidStorageData(stored_data) {
            return stored_data.hasOwnProperty("response");
        }
    }, {
        key: "isMethodAllowed",
        value: function isMethodAllowed(query_type, http_method, data_config) {
            return Adhara.app.allowedHttpMethods.indexOf(http_method) !== -1 || data_config.allowed_query_types && data_config.allowed_query_types.indexOf(query_type) !== -1;
        }
    }, {
        key: "isValidDataConfig",
        value: function isValidDataConfig(data_config) {
            return data_config instanceof Object && data_config.allowed_query_types instanceof Array && typeof data_config.url === "string" && data_config.allowed_query_types.length > 0 || data_config instanceof Object && data_config.batch_data_override instanceof Array;
        }
    }, {
        key: "isValidProcessor",
        value: function isValidProcessor(processor) {
            return processor instanceof Object && processor.success instanceof Function && processor.error instanceof Function;
        }
    }, {
        key: "isValidEntityConfig",
        value: function isValidEntityConfig(entity_config) {
            if (entity_config instanceof Object && this.isValidDataConfig(entity_config.data_config)) {
                if (entity_config.data_config.batch_data_override) {
                    return true;
                } else if (entity_config.data_config.allowed_query_types.indexOf('get') > -1 || entity_config.data_config.allowed_query_types.indexOf('get_list') > -1) {
                    if (this.isValidProcessor(entity_config.processor)) {
                        return true;
                    } else if (!AdharaView.isPrototypeOf(entity_config.view)) {
                        return false;
                    }
                }
                return true;
            }
            return false;
        }
    }, {
        key: "getHTTPMethod",
        value: function getHTTPMethod(query_type) {
            return query_type === "get_list" ? "get" : query_type;
        }
    }, {
        key: "getUniqueUrlForData",
        value: function getUniqueUrlForData(url, http_method, data) {
            return url + (data ? "?" + Object.keys(data).sort().map(function (key) {
                return key + "=" + data[key];
            }).join("&") : "");
        }
    }, {
        key: "remember",
        value: function remember(url, response, reuse, _url) {
            var _this31 = this;

            var expires = (reuse.timeout || this.config.reuse_timeout || 5 * 60 * 1000) + Date.now(); //5 minutes is the default timeout
            var _ = {
                expires: expires,
                url: _url || url
            };
            if (reuse.scope === "in_page") {
                _.page_name = Adhara.router.getCurrentPageName();
            }
            this.rem_que[url] = response; // hold response till dbPromise resolves
            return this.db_table.store({ url: url, response: response, _: _ }).then(function () {
                delete _this31.rem_que[url];
                return true;
            });
        }
    }, {
        key: "recall",
        value: function recall(data_url) {
            var _this32 = this;

            return new Promise(function (resolve, reject) {
                if (_this32.rem_que[data_url]) {
                    resolve(_this32.rem_que[data_url]);
                } else {
                    _this32.db_table.retrieve(data_url).then(function (data) {
                        if (!data) {
                            reject({ message: "No such key stored", code: 404 });
                        } else if (isNaN(data._.expires) || data._.expires < new Date().getTime()) {
                            reject({ message: "data has expired", code: 404 });
                        } else {
                            _this32.isValidStorageData(data) ? resolve(data.response) : reject({ message: "Invalid data " + data.response, code: 500 });
                        }
                    }, function (e) {
                        if (e.code === 8) {
                            reject({ message: "No such key stored", code: 404 });
                        } else {
                            reject({ message: "Unknown Error", code: 500 });
                        }
                    }); //.catch(reject);
                }
            });
        }
    }, {
        key: "remove",
        value: function remove(data_url) {
            return this.db_table.remove(data_url);
        }
    }, {
        key: "signalViewSuccess",
        value: function signalViewSuccess(query_type, entity_config, response_json, xhr) {
            Adhara.configUtils.getProcessor(entity_config).success(query_type, entity_config, response_json, xhr, true);
        }
    }, {
        key: "signalViewFailure",
        value: function signalViewFailure(query_type, entity_config, error, xhr) {
            Adhara.configUtils.getProcessor(entity_config).error(query_type, entity_config, error, xhr);
        }

        // rest of the public methods and local glue logic goes here ...

    }, {
        key: "enqueue",
        value: function enqueue(query_type, entity_config, data, options) {
            var _this33 = this;

            // AS OF NOW options can contain additional parameter 'caller_view'. Which ensures the caller view does get the result.
            if (!this.isValidEntityConfig(entity_config)) {
                throw new Error("Entity config passed to enqueue is invalid: " + JSON.stringify(entity_config, null, 4));
            }
            var data_config = Adhara.configUtils.getDataConfig(entity_config);
            if (data_config.hasOwnProperty("batch_data_override")) {
                // Map of url -> data
                this.handle_batch_enqueue(entity_config);
                return;
            }
            if (entity_config instanceof Array) {
                // Array of app configs -> array of responses
                // batch call type 2
                this.handle_bulk_config(query_type, entity_config, data, options);
                return;
            }
            // if it is not a get or get_list it will make it wait in a queue
            if (!(this.request_queue[data_config.url] instanceof Array)) {
                this.request_queue[data_config.url] = [];
            }
            if (['get', 'get_list'].indexOf(query_type) < 0) {
                if (options && options.consider_for_queueing !== false) {
                    options.consider_for_queueing = false; // would prevent recursive queueing of the same URL
                    this.request_queue[data_config.url].push({ entry_time: performance.now(), arg: [query_type, entity_config, data, options] });
                    if (this.request_queue[data_config.url].length > 1) {
                        // not the first entry, taken up later hence
                        return;
                    }
                }
            }

            //=========================================================================================================//
            //                  LOGISTICS END HERE, CORE ENQUEUE IS BELOW THIS COMMENT
            //=========================================================================================================//

            if (!query_type) query_type = data_config.default_query_type;
            var http_method = this.getHTTPMethod(query_type);
            if (!this.isMethodAllowed(query_type, http_method, data_config)) {
                // either globally off or off due to links configured for particular data configuration
                var failure_message = { message: "Unauthorized Request" };
                this.signalViewFailure(query_type, entity_config, failure_message, 405);
                return;
            }
            var reuse = data_config.reuse || {};
            if (reuse.enable !== false && this.config.default_reuse !== false && ['get', 'get_list'].indexOf(http_method) !== -1) {
                var unique_url = this.getUniqueUrlForData(data_config.url, http_method, data);
                var msc = function msc() {
                    //initiating call to Backend Service, and registering listeners for success and failure
                    _this33.makeServiceCall(data_config.url, http_method, data, Adhara.configUtils.getController(entity_config)).then(function (response_object) {
                        var response = Adhara.app.responseMiddleWare(entity_config, true, response_object.response, response_object.xhr);
                        _this33.signalViewSuccess(query_type, entity_config, response, response_object.xhr);
                        _this33.remember(unique_url, response, reuse, data_config._url);
                    }, function (error_response_object) {
                        _this33.signalViewFailure(query_type, entity_config, Adhara.app.responseMiddleWare(entity_config, false, error_response_object.error, error_response_object.xhr), error_response_object.xhr);
                    });
                };
                this.recall(unique_url).then(function (response) {
                    if (!(reuse instanceof Function) || reuse(response)) {
                        var xhr = new XMLHttpRequest();
                        _this33.signalViewSuccess(query_type, entity_config, response, xhr);
                    } else {
                        msc();
                    }
                }, function (err) {
                    msc();
                });
            } else {
                this.makeServiceCall(data_config.url, http_method, data, Adhara.configUtils.getController(entity_config)).then(function (response_object) {
                    _this33.signalViewSuccess(query_type, entity_config, Adhara.app.responseMiddleWare(entity_config, true, response_object.response, response_object.xhr), response_object.xhr);
                }, function (error_response_object) {
                    _this33.signalViewFailure(query_type, entity_config, Adhara.app.responseMiddleWare(entity_config, false, error_response_object.error, error_response_object.xhr), error_response_object.xhr);
                }).then(function () {
                    _this33.request_queue[data_config.url].shift();
                    if (_this33.request_queue[data_config.url].length > 0) {
                        if (performance.now() - _this33.request_queue[data_config.url][0]['entry_time'] <= 30000) {
                            _this33.enqueue.apply(_this33, _toConsumableArray(_this33.request_queue[data_config.url][0]['arg']));
                        } else {
                            _this33.request_queue[data_config.url] = [];
                        }
                    }
                });
            }
        }
    }, {
        key: "cleanUp",
        value: function cleanUp(current_page_name) {
            this.db_table.removeMultiple(function (url, response) {
                return response._.expires <= Date.now() || response._.page_name && response._.page_name !== current_page_name;
            });
        }
    }, {
        key: "initDependencies",
        value: function initDependencies() {
            var _this34 = this;

            Adhara.router.onRoute("DIPageChangeListener", function () {
                _this34.cleanUp(Adhara.router.getCurrentPageName());
            });
        }
    }]);

    return DataInterface;
}(StorageSelector.select());

/**------------------------------------------------------------------------------------------------------------------**/


var processor_helper = {
    on_success_common: function on_success_common(query_type, entity_config, response, xhr, pass_over) {
        // put here the common functionality that is shared among all the processors on success
        // like intimating the data to other processors of contexts with mutually same data_configs
        var data_config = Adhara.configUtils.getDataConfig(entity_config);
        if (!data_config) return false;
        var related_app_configs = Adhara.configUtils.getByDataConfig(data_config);
        for (var i = 0; i < related_app_configs.length; i++) {
            var processor = Adhara.configUtils.getProcessor(related_app_configs[i]);
            if (processor) {
                processor.success(query_type, related_app_configs[i], response, xhr, pass_over);
            }
        }
    },
    get_basic_processed_data: function get_basic_processed_data(query_type, entity_config, response, xhr) {
        var blob = Adhara.configUtils.getBlobClass(entity_config);
        var processed_data = void 0;
        if (query_type === "get_list") {
            processed_data = [];
            var _iteratorNormalCompletion16 = true;
            var _didIteratorError16 = false;
            var _iteratorError16 = undefined;

            try {
                for (var _iterator16 = response[Symbol.iterator](), _step16; !(_iteratorNormalCompletion16 = (_step16 = _iterator16.next()).done); _iteratorNormalCompletion16 = true) {
                    var datum = _step16.value;

                    processed_data.push(new blob(datum));
                }
            } catch (err) {
                _didIteratorError16 = true;
                _iteratorError16 = err;
            } finally {
                try {
                    if (!_iteratorNormalCompletion16 && _iterator16.return) {
                        _iterator16.return();
                    }
                } finally {
                    if (_didIteratorError16) {
                        throw _iteratorError16;
                    }
                }
            }
        } else {
            processed_data = new blob(response);
        }
        return processed_data;
    }
};

var Processor = {

    fallback: {
        success: function success(query_type, entity_config, response, xhr, pass_over) {
            var processed_data = processor_helper.get_basic_processed_data(query_type, entity_config, response, xhr);
            var view = Adhara.configUtils.getViewInstance(entity_config);
            // move on to the functionality common among all the processors
            if (pass_over === true && !processor_helper.on_success_common(query_type, entity_config, response, xhr, false) && view) {
                view.handleDataChange(processed_data);
            } else if (view) {
                view.handleDataChange(processed_data);
            }
        },
        error: function error(query_type, entity_config, _error4, xhr) {
            var processed_data = { error: _error4, response_code: xhr.status };
            // let processed_data = processor_helper.get_basic_processed_data(query_type, entity_config, error, xhr);
            //TODO is it required to create error response as a blob?
            var view = Adhara.configUtils.getViewInstance(entity_config);
            if (view) {
                view.handleDataError(processed_data);
            }
        }
    }

};
// register with this object all the processors for the data as per all the contexts
/**------------------------------------------------------------------------------------------------------------------**/

var WebSocket = function () {
    function WebSocket(web_socket_config) {
        _classCallCheck(this, WebSocket);

        var socket = this.socket = io.connect(web_socket_config.url || "/socket", {
            resource: web_socket_config.resource || "/"
        });
        socket.on('connect_failed', function (data) {
            return console.log('connect_failed', data);
        });
        socket.on('connecting', function (data) {
            return console.log('connecting', data);
        });
        socket.on('disconnect', function (data) {
            return console.log('disconnect', data);
        });
        socket.on('error', function (reason) {
            return console.log('error', reason);
        });
        socket.on('reconnect_failed', function (data) {
            return console.log('reconnect_failed', data);
        });
        socket.on('reconnect', function (data) {
            return console.log('reconnect', data);
        });
        socket.on('reconnecting', function (data) {
            return console.log('reconnecting', data);
        });
    }

    _createClass(WebSocket, [{
        key: "listen",
        value: function listen() {
            var _this35 = this;

            // instance scope method
            var config = Adhara.app.config;

            var _loop5 = function _loop5(app_conf) {
                if (config.hasOwnProperty(app_conf)) {
                    var data_config = Adhara.configUtils.getDataConfig(app_conf);
                    if (!data_config.socket_tag) return "continue";
                    _this35.socket.on(data_config.socket_tag, function (data) {
                        Adhara.configUtils.getProcessor(app_conf)(data);
                    });
                }
            };

            for (var app_conf in config) {
                var _ret5 = _loop5(app_conf);

                if (_ret5 === "continue") continue;
            }
        }
    }]);

    return WebSocket;
}();

WebSocket.listen = function (web_socket_config) {
    // static scope method
    var ws = new WebSocket(web_socket_config);
    ws.listen();
};
/**------------------------------------------------------------------------------------------------------------------**/

var AdharaApp = function () {
    function AdharaApp() {
        _classCallCheck(this, AdharaApp);
    }

    _createClass(AdharaApp, [{
        key: "getEntityConfigFromContext",


        /**
         * @function
         * @instance
         * @return {Object} Adhara style entity config from crude entity config
         * */
        value: function getEntityConfigFromContext(context) {
            var allowed_query_types = context.data_config.allowed_query_types ? context.data_config.allowed_query_types.slice() : [];
            var data_config = void 0;
            if (context.data_config.hasOwnProperty("batch_data_override")) {
                data_config = {
                    batch_data_override: context.data_config.batch_data_override.map(function (batch_data_config) {
                        if (!batch_data_config.reuse) {
                            batch_data_config.reuse = {};
                        }
                        return {
                            _url: batch_data_config.url,
                            url: batch_data_config.url,
                            query_type: batch_data_config.query_type,
                            identifier: batch_data_config.identifier,
                            reuse: {
                                enable: batch_data_config.reuse.enable,
                                timeout: batch_data_config.reuse.timeout,
                                scope: batch_data_config.reuse.scope,
                                handler: batch_data_config.reuse.handler
                            },
                            blob: batch_data_config.blob
                        };
                    })
                };
            } else {
                if (!context.data_config.reuse) {
                    context.data_config.reuse = {};
                }
                data_config = {
                    _url: context.data_config.url,
                    url: context.data_config.url,
                    allowed_query_types: allowed_query_types,
                    default_query_type: context.data_config.default_query_type || allowed_query_types[0],
                    socket_tag: context.data_config.socket_tag,
                    reuse: {
                        enable: context.data_config.reuse.enable,
                        timeout: context.data_config.reuse.timeout,
                        scope: context.data_config.reuse.scope,
                        handler: context.data_config.reuse.handler
                    },
                    blob: context.data_config.blob
                };
            }
            var processor = {
                success: context.processor && context.processor.success || Processor.fallback.success,
                error: context.processor && context.processor.error || Processor.fallback.error
            };
            var controller = {
                get: context.controller && context.controller.get || Adhara.restAPI.get,
                put: context.controller && context.controller.put || Adhara.restAPI.put,
                post: context.controller && context.controller.post || Adhara.restAPI.post,
                delete: context.controller && context.controller.delete || Adhara.restAPI.delete
            };
            return {
                data_config: data_config,
                view: context.view,
                controller: controller,
                processor: processor
            };
        }

        /**
         * @function
         * @instance
         * @return {Object} Adhara style entity config
         * */

    }, {
        key: "getEntityConfig",
        value: function getEntityConfig(context_name) {
            return this.getEntityConfigFromContext(this.config[context_name]);
        }

        /**
         * @getter
         * @instance
         * @returns {Object} Custom global view configurations
         * */

    }, {
        key: "responseMiddleWare",


        /**
         * @function
         * @instance
         * @description modify the response even before being utilized by the DataInterface
         * */
        value: function responseMiddleWare(entity_config, success, response, xhr) {
            return response;
        }

        /**
         * @function
         * @instance
         * @description modify the data with respect to the URL and method name and
         * return formatted data to be posted or queried to/from the server
         * */

    }, {
        key: "requestMiddleWare",
        value: function requestMiddleWare(url, method_name, data) {
            return data;
        }

        /**
         * @function
         * @instance
         * @param {String} title - toast message title
         * @param {String} content - toast message content
         * @param {String} type - toast message type. can take the values "success"|"error"|"info".
         * */

    }, {
        key: "toast",
        value: function toast(title, content, type) {
            return AdharaDefaultToaster.make(title, content, type);
        }
    }, {
        key: "is_debug_mode",


        /**
         * @function
         * @instance
         * @return {Boolean} Whether the app is being accessed in development mode or in production mode
         * */
        get: function get() {
            return ["localhost", "127.0.0.1", "0.0.0.0"].indexOf(window.location.hostname) !== -1 || window.location.hostname.indexOf("192.168.") !== -1;
        }

        /**
         * @function
         * @instance
         * @return {String} App name
         * */

    }, {
        key: "name",
        get: function get() {
            return "Adhara App";
        }

        /**
         * @function
         * @instance
         * @return {Object} App name in detail. Like first and last name
         * */

    }, {
        key: "detailedName",
        get: function get() {
            return {
                first: "Adhara",
                last: "App"
            };
        }

        /**
         * @function
         * @instance
         * @return {String} Tag line
         * */

    }, {
        key: "tagLine",
        get: function get() {}

        /**
         * @function
         * @instance
         * @return {AdharaView} Container view class
         * */

    }, {
        key: "containerView",
        get: function get() {}

        /**
         * @function
         * @instance
         * @return {Object} Adhara style app config
         * */

    }, {
        key: "config",
        get: function get() {
            return {};
        }
    }, {
        key: "customViewConfig",
        get: function get() {
            return {
                fetching_data: {
                    AdharaListView: "Fetching Data..."
                },
                no_data: {
                    AdharaListView: "No Data Available"
                }
            };
        }

        /**
         * @function
         * @instance
         * @returns {Array<String>} list of http methods to be allowed by the application.
         * @description This getter can be configured to return allowed methods based on the current network state.
         * Say if offline, it can be configured to just return `["get"]` method which will restrict DataInterface from making
         * other service API calls such as "post", "delete", etc...
         * */

    }, {
        key: "allowedHttpMethods",
        get: function get() {
            return ['get', 'post', 'put', 'delete']; // all available API methods
            // offline mode will switch a few of these off (post, put and delete)
        }

        /**
         * @function
         * @instance
         * @return {AdharaRouterConfiguration} Adhara style routing config
         * */

    }, {
        key: "routerConfiguration",
        get: function get() {
            return {
                routes: [],
                on_route_listeners: {},
                middlewares: []
            };
        }

        /**
         * @function
         * @instance
         * @returns {String} A css selector in which app is to be rendered.
         * */

    }, {
        key: "DOMSelector",
        get: function get() {
            return "app"; //=> search DOM for `<app></app>`
        }

        /**
         * @function
         * @instance
         * @returns {String} API Server URL. Either the base path or a full url till base path.
         * */

    }, {
        key: "apiServerURL",
        get: function get() {
            return "/api";
        }

        /**
         * @function
         * @instance
         * @returns {Object} WebSocket config object.
         * sample...
         * {
         *  url: "sub.domain.com:9081"
         * }
         * */

    }, {
        key: "webSocketConfig",
        get: function get() {}

        /**
         * @getter
         * @instance
         * @returns {Object} DB Configuration which includes DB names and DB Schematics for Client Storage ( for all the object stores )
         * */

    }, {
        key: "DBConfig",
        get: function get() {
            return [{
                id: "default",
                name: "Adhara-app_db",
                version: 1,
                schema: {
                    http_cache: {
                        keyPath: "url"
                    }
                }
            }];
        }
    }, {
        key: "_diConfig",
        get: function get() {
            return Object.assign({
                http_cache_table: "http_cache",
                default_reuse: true,
                reuse_timeout: 5 * 60 * 1000
            }, this.DIConfig);
        }

        /**
         * @getter
         * @instance
         * @returns {Object} Data interface related configuration
         * */

    }, {
        key: "DIConfig",
        get: function get() {
            return {};
        }

        /**
         * @getter
         * @instance
         * @returns {Object<String, Object>} scopes that to be used in the app.
         * @description Map of scopes that can be accessed across the application.
         * Scopes can vary from environment to environment. Like in Node.JS vs in Browser JS.
         * */

    }, {
        key: "scopes",
        get: function get() {
            return {
                global: window // the global scope
            };
        }
    }, {
        key: "i18n_key_map",
        get: function get() {
            return {
                "list_view.navigation.previous": "Previous",
                "list_view.navigation.next": "Next"
            };
        }
    }]);

    return AdharaApp;
}();

/**------------------------------------------------------------------------------------------------------------------**/
/**
 * @class
 * @classdesc a class that handles everything related to the website
 * */


var Adhara = null;

(function () {
    var AdharaBase = function () {
        function AdharaBase() {
            _classCallCheck(this, AdharaBase);

            this.container = null;
            this.always_active_views = [];
            this.active_views = [];
            this.router = AdharaRouter;
            this.toast = Toast;
        }

        _createClass(AdharaBase, [{
            key: "init",
            value: function init(app) {
                callOnInitListeners();
                if (app) {
                    this.app = new app();
                    this.dataInterface = new DataInterface();
                    this.restAPI = RestAPI;
                    this.i18n = new Internationalize(Adhara.app.i18n_key_map);
                    this.createShortcuts();
                    this.performSystemChecks();
                    this.createContainer();
                    if (this.app.webSocketConfig) {
                        WebSocket.listen(this.app.webSocketConfig);
                    }
                } else {
                    this.router.route();
                }
            }
        }, {
            key: "createShortcuts",
            value: function createShortcuts() {
                var _this36 = this;

                //Creating a view_name vs context_name map
                this.view_context = {};
                loop(this.app.config, function (context_name, configuration) {
                    if (configuration.hasOwnProperty("view") && AdharaView.isPrototypeOf(configuration.view)) {
                        _this36.view_context[configuration.view.name] = context_name;
                    }
                });
            }
        }, {
            key: "getViewContext",
            value: function getViewContext(viewInstance) {
                return this.view_context[viewInstance.constructor.name];
            }
        }, {
            key: "performSystemChecks",
            value: function performSystemChecks() {
                /*// verify the controller for if it has support for all the API methods (for ease of Dev)
                for(let i = 0; i < Adhara.app.allowedHttpMethods.length; i++){
                    if(typeof Controller[Adhara.app.allowedHttpMethods[i] + "Data"] !== 'function'){
                        throw new Error(Adhara.app.allowedHttpMethods[i] + " api method is not registered with the controller!");
                    }
                }*/
            }
        }, {
            key: "createContainer",
            value: function createContainer() {
                this.router.configure(this.app.routerConfiguration);
                this.router.listen();
                if (this.app.containerView) {
                    this.container = new this.app.containerView();
                    this.container.onViewRendered(this.router.route);
                    this.always_active_views = this.container.subViews.map(function (subView) {
                        return Adhara.getView(subView);
                    });
                    this.always_active_views.push(this.container);
                    this.clearActiveViews();
                    Adhara.createView(this.container);
                } else {
                    this.clearActiveViews();
                    this.router.route();
                }
            }
        }, {
            key: "addToActiveViews",
            value: function addToActiveViews(viewInstance) {
                if (!this.isActiveView(viewInstance)) {
                    this.active_views.push(viewInstance);
                }
            }
        }, {
            key: "getActiveView",
            value: function getActiveView(classReference) {
                var active_view = this.active_views.filter(function (view_instance) {
                    return view_instance.constructor === classReference;
                })[0];
                if (!active_view) {
                    active_view = this.always_active_views.filter(function (view_instance) {
                        return view_instance.constructor === classReference;
                    })[0];
                }
                return active_view;
            }
        }, {
            key: "isActiveView",
            value: function isActiveView(viewInstance) {
                return this.active_views.indexOf(viewInstance) !== -1 || this.always_active_views.indexOf(viewInstance) !== -1;
            }
        }, {
            key: "clearActiveViews",
            value: function clearActiveViews() {
                this.active_views = [];
            }
        }, {
            key: "closeActiveViews",
            value: function closeActiveViews() {
                var _iteratorNormalCompletion17 = true;
                var _didIteratorError17 = false;
                var _iteratorError17 = undefined;

                try {
                    for (var _iterator17 = this.active_views[Symbol.iterator](), _step17; !(_iteratorNormalCompletion17 = (_step17 = _iterator17.next()).done); _iteratorNormalCompletion17 = true) {
                        var active_view = _step17.value;

                        active_view.destroy();
                    }
                } catch (err) {
                    _didIteratorError17 = true;
                    _iteratorError17 = err;
                } finally {
                    try {
                        if (!_iteratorNormalCompletion17 && _iterator17.return) {
                            _iterator17.return();
                        }
                    } finally {
                        if (_didIteratorError17) {
                            throw _iteratorError17;
                        }
                    }
                }

                this.clearActiveViews();
            }
        }, {
            key: "templateEngine",
            get: function get() {
                return {
                    name: "Handlebars",
                    helpersHandler: TemplateEngineHelpers
                };
            }
        }, {
            key: "viewHierarchy",
            get: function get() {
                return [AdharaContainerView, AdharaDialogView, AdharaTemplateView, AdharaCardView, AdharaGridView, AdharaListView, AdharaFormView, AdharaView];
            }
        }]);

        return AdharaBase;
    }();

    Adhara = new AdharaBase();

    //View class handling
    var view_instances = {};

    //Single ton views
    Adhara.addViewToInstances = function (instance) {
        view_instances[instance.constructor.name] = instance;
    };

    //Query singleton views
    Adhara.getView = function (viewClass, parentViewInstance) {
        if (!viewClass) {
            throw new Error("invalid view class");
        }
        var active_view = Adhara.getActiveView(viewClass instanceof Function ? viewClass : viewClass.constructor);
        if (active_view) {
            return active_view;
        } else if (parentViewInstance) {
            return new viewClass(parentViewInstance);
        } else {
            throw new Error("parent view is required to construct a view. Failed to create view: " + viewClass.constructor.name);
        }
    };

    //Create a view instance
    Adhara.createView = function (adhara_view_instance, parentViewInstance) {
        adhara_view_instance.create(parentViewInstance);
    };

    //On route listener
    Adhara.onRoute = function (view_class) {
        Adhara.closeActiveViews();
        if (!Adhara.container.isActive()) {
            return Adhara.container.create();
        }
        Adhara.createView(Adhara.getView(view_class, Adhara.container));
    };

    var on_init_listeners = [registerAdharaUtils];

    Adhara.onInit = function (fn) {
        on_init_listeners.push(fn);
    };

    Adhara.lightReload = function () {
        Adhara.container ? Adhara.container.refresh() : Adhara.router.route();
    };

    function callOnInitListeners() {
        var _iteratorNormalCompletion18 = true;
        var _didIteratorError18 = false;
        var _iteratorError18 = undefined;

        try {
            for (var _iterator18 = on_init_listeners[Symbol.iterator](), _step18; !(_iteratorNormalCompletion18 = (_step18 = _iterator18.next()).done); _iteratorNormalCompletion18 = true) {
                var on_init_listener = _step18.value;

                on_init_listener();
            }
        } catch (err) {
            _didIteratorError18 = true;
            _iteratorError18 = err;
        } finally {
            try {
                if (!_iteratorNormalCompletion18 && _iterator18.return) {
                    _iterator18.return();
                }
            } finally {
                if (_didIteratorError18) {
                    throw _iteratorError18;
                }
            }
        }
    }
})();
//# sourceMappingURL=adhara.combined.es5.js.map
